from .base import BaseDetector
from .hr_patchcore import HRPatchCore
from .hr_fastflow import HRFastFlow
from .hr_vitad import HRVitAD
from .hr_realnet import HRRealNet
from .hr_destseg import HRDesTSeg
from .hr_rdpp import HRRDPP
from .hr_padim import HRPaDiM



from typing import Callable, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


def conv3x3(
    in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1
) -> nn.Conv2d:
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


def make_layer(block, inplanes, planes, blocks, stride=1, norm_layer=None):
    if norm_layer is None:
        norm_layer = nn.BatchNorm2d
    downsample = None
    if stride != 1 or inplanes != planes * block.expansion:
        downsample = nn.Sequential(
            conv1x1(inplanes, planes * block.expansion, stride),
            norm_layer(planes * block.expansion),
        )

    layers = []
    layers.append(block(inplanes, planes, stride, downsample, norm_layer=norm_layer))
    inplanes = planes * block.expansion
    for _ in range(1, blocks):
        layers.append(block(inplanes, planes, norm_layer=norm_layer))

    return nn.Sequential(*layers)


def l2_normalize(input, dim=1, eps=1e-12):
    denom = torch.sqrt(torch.sum(input**2, dim=dim, keepdim=True))
    return input / (denom + eps)


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


def get_norm_layer(norm: str):
    norm = {
        "BN": nn.BatchNorm2d,
        "LN": nn.LayerNorm,
    }[norm.upper()]
    return norm


def get_act_layer(act: str):
    act = {
        "relu": nn.ReLU,
        "relu6": nn.ReLU6,
        "swish": nn.SiLU,
        "mish": nn.Mish,
        "leaky_relu": nn.LeakyReLU,
        "sigmoid": nn.Sigmoid,
        "gelu": nn.GELU,
    }[act.lower()]
    return act


class ConvNormAct2d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding="same",
        dilation=1,
        groups=1,
        conv_kwargs=None,
        norm_layer=None,
        norm_kwargs=None,
        act_layer=None,
        act_kwargs=None,
    ):
        super(ConvNormAct2d, self).__init__()

        conv_kwargs = {}
        if norm_layer:
            conv_kwargs["bias"] = False
        if padding == "same" and stride > 1:
            # if kernel_size is even, -1 is must
            padding = (kernel_size - 1) // 2

        self.conv = self._build_conv(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            conv_kwargs,
        )
        self.norm = None
        if norm_layer:
            norm_kwargs = {}
            self.norm = get_norm_layer(norm_layer)(
                num_features=out_channels, **norm_kwargs
            )
        self.act = None
        if act_layer:
            act_kwargs = {}
            self.act = get_act_layer(act_layer)(**act_kwargs)

    def _build_conv(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
        dilation,
        groups,
        conv_kwargs,
    ):
        return nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            **conv_kwargs,
        )

    def forward(self, x):
        x = self.conv(x)
        if self.norm:
            x = self.norm(x)
        if self.act:
            x = self.act(x)
        return x


class ASPP(nn.Module):
    def __init__(self, input_channels, output_channels, atrous_rates):
        super(ASPP, self).__init__()
        modules = []
        modules.append(
            nn.Sequential(
                nn.AdaptiveAvgPool2d(1),
                ConvNormAct2d(
                    input_channels,
                    output_channels,
                    kernel_size=1,
                    norm_layer="BN",
                    act_layer="RELU",
                ),
            )
        )
        for atrous_rate in atrous_rates:
            conv_norm_act = ConvNormAct2d
            modules.append(
                conv_norm_act(
                    in_channels=input_channels,
                    out_channels=output_channels,
                    kernel_size=1 if atrous_rate == 1 else 3,
                    padding=0 if atrous_rate == 1 else atrous_rate,
                    dilation=atrous_rate,
                    norm_layer="BN",
                    act_layer="RELU",
                )
            )

        self.aspp_feature_extractors = nn.ModuleList(modules)
        self.aspp_fusion_layer = ConvNormAct2d(
            (1 + len(atrous_rates)) * output_channels,
            output_channels,
            kernel_size=3,
            norm_layer="BN",
            act_layer="RELU",
        )

    def forward(self, x):
        res = []
        for aspp_feature_extractor in self.aspp_feature_extractors:
            res.append(aspp_feature_extractor(x))
        res[0] = F.interpolate(
            input=res[0], size=x.shape[2:], mode="bilinear", align_corners=False
        )  # resize back after global-avg-pooling layer
        res = torch.cat(res, dim=1)
        res = self.aspp_fusion_layer(res)
        return res




import logging
from datetime import datetime
import random
import os
import torch
import numpy as np
from collections.abc import Mapping
import shutil


def map_func(storage, location):
    return storage.cuda()


def load_pertrain_weights(pertrain_path,model):
    pertrain_state_dict = torch.load(pertrain_path,map_location=map_func)
    if 'state_dict' in list(pertrain_state_dict.keys()):
        pertrain_state_dict = pertrain_state_dict['state_dict']
        pertrain_state_dict_= {}
        for key in pertrain_state_dict:
            if key.startswith("module"):
                key_new=key.replace("module.","")
            else:
                key_new=key
            pertrain_state_dict_[key_new]=pertrain_state_dict[key]
        pertrain_state_dict=pertrain_state_dict_
    model_state_dict = model.state_dict()
    match_keys = {}
    unexpected_keys = []
    for key in pertrain_state_dict:
        if key in model_state_dict.keys():
            if pertrain_state_dict[key].size() == model_state_dict[key].size():
                match_keys[key] = pertrain_state_dict[key]
        else:
            unexpected_keys.append(key)
    missing_keys = [key for key in model_state_dict.keys() if key not in match_keys]
    for key in match_keys:
        model_state_dict[key].data.copy_(match_keys[key])
    return missing_keys,unexpected_keys


def summary_model(model,logger):
    logger.info("************* model summary *************")
    for name,child in model.named_children():
        logger.info('{}: {} param'.format(name,sum([param.numel()for param in child.parameters()])))
    total_num=sum([param.numel() for param in model.parameters()])
    frozen_num = sum([param.numel()  for layer in model.frozen_layers for param in getattr(model,layer).parameters()])
    logger.info("************* model summary *************")
    logger.info('total: {} param'.format(total_num))
    logger.info('frozen: {} param'.format(frozen_num))
    logger.info('trainable: {} param'.format(total_num-frozen_num))
    logger.info("************* model summary *************")


def set_seed(seed):
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    random.seed(seed)


def get_current_time():
    current_time = datetime.now().strftime("%Y-%m-%d %H:%M")
    return current_time


def ema(source, target, decay):
    source_dict = source.state_dict()
    target_dict = target.state_dict()

    for key in source_dict.keys():
        target_dict[key].data.copy_(target_dict[key].data * decay +
                                    source_dict[key].data * (1 - decay))


def to_device(input, device="cuda", dtype=None):
    """Transfer data between devidces"""

    if "image" in input:
        input["image"] = input["image"].to(dtype=dtype)

    def transfer(x):
        if torch.is_tensor(x):
            return x.to(device=device)
        elif isinstance(x, list):
            return [transfer(_) for _ in x]
        elif isinstance(x, Mapping):
            return type(x)({k: transfer(v) for k, v in x.items()})
        else:
            return x

    return {k: transfer(v) for k, v in input.items()}


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, length=0):
        self.length = length
        self.reset()

    def reset(self):
        if self.length > 0:
            self.history = []
        else:
            self.count = 0
            self.sum = 0.0
        self.val = 0.0
        self.avg = 0.0

    def update(self, val, num=1):
        if self.length > 0:
            # currently assert num==1 to avoid bad usage, refine when there are some explict requirements
            assert num == 1
            self.history.append(val)
            if len(self.history) > self.length:
                del self.history[0]

            self.val = self.history[-1]
            self.avg = np.mean(self.history)
        else:
            self.val = val
            self.sum += val * num
            self.count += num
            self.avg = self.sum / self.count


from typing import List
import numpy
import numpy as np
import torch
import logging
from torch.utils.data import DataLoader
import torch.nn.functional as F
from scipy.ndimage import gaussian_filter

from hiad.utils.split_and_gather import LRPatch
from hiad.datasets.syn_dataset import AnomalySynDataset
from hiad.datasets.patch_dataset import PatchDataset
from hiad.syn import *
from .base import BaseDetector
from .rdpp import resnet
from .rdpp import de_resnet
from .rdpp.utils.utils_train import MultiProjectionLayer, Revisit_RDLoss, loss_fucntion


class HRRDPP(BaseDetector):

    def __init__(self,
                 patch_size: int, # base
                 backbone_name: str,
                 layers_to_extract_from: list,
                 synthesizer,
                 epoch: int,
                 proj_lr: float,
                 distill_lr: float,
                 weight_proj: float,
                 logger: logging.Logger, #base
                 device: torch.device, #base
                 seed: int = 0, #base
                 log_per_steps: int = 10,
                 eval_per_steps: int= 100,
                 fusion_weights = None,
                 early_stop_epochs = -1,
                 **kwargs):

        super().__init__(patch_size, device, fusion_weights, logger, seed, early_stop_epochs)

        self.backbone_name = backbone_name
        self.epoch = epoch
        self.proj_lr = proj_lr
        self.distill_lr = distill_lr
        self.weight_proj =  weight_proj
        self.synthesizer = synthesizer
        self.log_per_steps = log_per_steps
        self.eval_per_steps = eval_per_steps

        self.layers_to_extract_from = layers_to_extract_from
        self.synthesizer = eval(self.synthesizer.name)(**self.synthesizer.kwargs)

        self.encoder, self.bn = resnet._BACKBONES[self.backbone_name](pretrained=True,
                                                                      layers_to_extract_from=self.layers_to_extract_from)
        self.encoder = self.encoder.to(device)
        self.bn = self.bn.to(device)
        self.encoder.eval()

        self.decoder = de_resnet._BACKBONES[self.backbone_name](pretrained=False, layers_to_extract_from=self.layers_to_extract_from)
        self.decoder = self.decoder.to(device)

        self.proj_layer = MultiProjectionLayer(base=64, layers_to_extract_from=self.layers_to_extract_from).to(device)

        self.max_anomaly_score = None
        self.min_anomaly_score = None

    def create_dataset(self, patches: List[LRPatch], training: bool, task_name: str ):
        if not training:
            dataset = PatchDataset(patches=patches, training=training, task_name=task_name)
        else:
            dataset = AnomalySynDataset(patches=patches, synthesizer=self.synthesizer, training=training, task_name = task_name)
        return dataset

    @torch.no_grad()
    def embedding(self, input_tensor: torch.Tensor ) -> List[torch.Tensor]:
        input_tensor = input_tensor.to(self.device)
        outputs = self.encoder(input_tensor)
        return outputs

    def rdpp_emb(self, samples):
        normal_image = {'image': samples['gt_image']}
        noise_image =  {'image': samples['image']}
        for key in samples:
            if key.startswith('gt_low_resolution_index') or key.startswith('gt_low_resolution_image'):
                normal_image[key.replace('gt_', '')] = samples[key]

            if key.startswith('low_resolution_index') or key.startswith('low_resolution_image'):
                noise_image[key] = samples[key]

        features, noise_features = self.get_multi_resolution_fusion_embeddings(normal_image), self.get_multi_resolution_fusion_embeddings(noise_image)
        return features, noise_features


    def train_step(self,
                   train_dataloader: DataLoader,
                   task_name: str,
                   checkpoint_path: str,
                   val_dataloader: DataLoader = None,
                   evaluators = None,
                   ) -> bool:

        self.logger.info(f"RD++: {task_name} start training ...")
        proj_loss = Revisit_RDLoss()

        optimizer_proj = torch.optim.Adam(list(self.proj_layer.parameters()), lr=self.proj_lr, betas=(0.5, 0.999))
        optimizer_distill = torch.optim.Adam(list(self.decoder.parameters()) + list(self.bn.parameters()), lr=self.distill_lr,
                                             betas=(0.5, 0.999))

        best_metrics = {}
        global_step = 0

        for epoch in range(self.epoch):

            loss_list = []
            for data in train_dataloader:
                torch.cuda.empty_cache()
                self.bn.train()
                self.proj_layer.train()
                self.decoder.train()

                global_step += 1

                features, noise_features = self.rdpp_emb(data)

                (feature_space_noise, feature_space) = self.proj_layer(features, features_noise=noise_features)

                L_proj = proj_loss(noise_features, feature_space_noise, feature_space)
                outputs = self.decoder(self.bn(feature_space))  # bn(inputs))
                L_distill = loss_fucntion(features, outputs)

                loss = L_distill + self.weight_proj * L_proj
                loss.backward()

                if global_step % 2 == 0:
                    optimizer_proj.step()
                    optimizer_distill.step()
                    optimizer_proj.zero_grad()
                    optimizer_distill.zero_grad()

                loss_list.append(loss.item())

                if global_step % self.log_per_steps == 0 or global_step % len(train_dataloader) == 0:
                    self.logger.info('epoch [{}/{}], step [{}], loss:{:.4f}'.format(epoch + 1, self.epoch, global_step,
                                                                                    np.mean(loss_list)))

                if global_step % self.eval_per_steps == 0 and val_dataloader is not None:
                    best_metrics = self.val_step(val_dataloader, evaluators, checkpoint_path, best_metrics)
                    if self.early_stop_cur == 0:
                        return True

        return True if best_metrics != {} else False


    @torch.no_grad()
    def inference_step(self,
                   test_dataloader: DataLoader,
                   task_name: str,
                   )-> List[numpy.ndarray]:

        self.bn.eval()
        self.decoder.eval()
        self.proj_layer.eval()

        preds = []
        with torch.no_grad():
            for data in test_dataloader:
                B = data['image'].shape[0]
                inputs = self.get_multi_resolution_fusion_embeddings(data)
                features = self.proj_layer(inputs)
                outputs = self.decoder(self.bn(features))

                inputs = [torch.split(input, 1) for input in inputs]
                outputs = [torch.split(output, 1) for output in outputs]

                for index in range(B):
                    input_ = [input[index] for input in inputs]
                    output_ = [output[index] for output in outputs]
                    anomaly_map, _ = self.cal_anomaly_map(input_, output_, self.patch_size, amap_mode='a')
                    anomaly_map = gaussian_filter(anomaly_map, sigma=4)
                    preds.append(np.expand_dims(anomaly_map, axis=0))

        preds = np.concatenate(preds, axis=0)

        if task_name != 'val':
            preds = self.patch_post_processing(preds)

        preds = np.split(preds, indices_or_sections=preds.shape[0])
        preds = [pred.squeeze(axis=0) for pred in preds]
        return preds


    def save_checkpoint(self, checkpoint_path: str):
        torch.save({'bn': self.bn.state_dict(), 'decoder': self.decoder.state_dict(),
                    'proj_layer': self.proj_layer.state_dict(),
                    'max_anomaly_score': self.max_anomaly_score, 'min_anomaly_score': self.min_anomaly_score,
                    'fusion_weights': self.fusion_weights},
                    checkpoint_path)


    def load_checkpoint(self, checkpoint_path: str):
        state_dict = torch.load(checkpoint_path, map_location=self.device)
        self.bn.load_state_dict(state_dict['bn'])
        self.decoder.load_state_dict(state_dict['decoder'])
        self.proj_layer.load_state_dict(state_dict['proj_layer'])
        self.min_anomaly_score = state_dict['min_anomaly_score']
        self.max_anomaly_score = state_dict['max_anomaly_score']
        self.fusion_weights = state_dict['fusion_weights'] if 'fusion_weights' in state_dict else None


    def cal_anomaly_map(self, fs_list, ft_list, out_size= (224, 224), amap_mode='mul'):
        if amap_mode == 'mul':
            anomaly_map = np.ones((out_size[1], out_size[0]))
        else:
            anomaly_map = np.zeros((out_size[1], out_size[0]))

        a_map_list = []
        for i in range(len(ft_list)):
            fs = fs_list[i]
            ft = ft_list[i]

            a_map = 1 - F.cosine_similarity(fs, ft)
            a_map = torch.unsqueeze(a_map, dim=1)
            a_map = F.interpolate(a_map, size=(out_size[1], out_size[0]), mode='bilinear', align_corners=True)
            a_map = a_map[0, 0, :, :].to('cpu').detach().numpy()
            a_map_list.append(a_map)
            if amap_mode == 'mul':
                anomaly_map *= a_map
            else:
                anomaly_map += a_map

        return anomaly_map, a_map_list



from typing import List
import torch
from torchvision import transforms
from torch.utils.data import Dataset
import numpy

from hiad.utils.split_and_gather import LRPatch


class PatchDataset(Dataset):

    def __init__(
        self,
        patches: List[LRPatch],
        training: bool,
        task_name: str,
        normalize_mean: List[float] = None,
        normalize_std: List[float] = None,
        **kwargs,
    ):

        super().__init__()
        self.patches = patches
        self.training = training
        self.task_name = task_name

        if normalize_mean is None:
            normalize_mean = [0.485, 0.456, 0.406]

        if normalize_std is None:
            normalize_std = [0.229, 0.224, 0.225]

        self.transform_img = [
            transforms.ToTensor(),
            transforms.Normalize(mean=normalize_mean, std=normalize_std),
        ]

        self.transform_mask = [
            transforms.ToTensor()
        ]

        self.transform_img = transforms.Compose(self.transform_img)
        self.transform_mask = transforms.Compose(self.transform_mask)

    def __getitem__(self, idx):

        patch = self.patches[idx]
        image = patch.image.astype(numpy.uint8)
        image = self.transform_img(image)

        if patch.mask is not None:
            mask = self.transform_mask(patch.mask.astype(numpy.uint8)).squeeze(0)
            mask [mask!=0] = 1
        else:
            mask = torch.zeros(image.shape[1:])

        item = {
            "image": image,
            "mask": mask,
        }

        if patch.clsname is not None:
            item.update({"clsname": patch.clsname})

        if patch.label is not None:
            item.update({"label": patch.label})

        if patch.low_resolution_images is not None:
            for i, (low_resolution_image, low_resolution_index) in enumerate(zip(patch.low_resolution_images, patch.low_resolution_indexes)):
                low_resolution_image = low_resolution_image.astype(numpy.uint8)
                low_resolution_image = self.transform_img(low_resolution_image)
                item[f'low_resolution_image_{i}'] = low_resolution_image
                item[f'low_resolution_index_{i}'] = str(low_resolution_index)
        return item

    def __len__(self):
        return len(self.patches)


from .unets import EncoderUNetModel, UNetModel
from hiad.detectors.realnet.utils.misc_helper import load_pertrain_weights


def create_diffusion_unet(
        image_size,
        classifier_width,
        classifier_depth,
        learn_sigma,
        attention_resolutions,
        dropout,
        channel_mult,
        class_number,
        num_heads,
        num_heads_upsample,
        num_head_channels,
        use_scale_shift_norm,
        resblock_updown,
        use_fp16,
        pertrain_path,
):
    unet = UNetModel(       image_size=image_size,
                            in_channels=3,
                            model_channels=classifier_width,
                            out_channels=(6 if learn_sigma else 3),
                            num_res_blocks=classifier_depth,
                            attention_resolutions=attention_resolutions,
                            dropout=dropout,
                            channel_mult=channel_mult,
                            num_classes=class_number,
                            num_heads=num_heads,
                            use_fp16=use_fp16,
                            num_heads_upsample=num_heads_upsample,
                            num_head_channels=num_head_channels,
                            use_scale_shift_norm=use_scale_shift_norm,
                            resblock_updown=resblock_updown
                        )

    if pertrain_path is not None:
        missing_keys, unexpected_keys = load_pertrain_weights(pertrain_path,unet)
        # print(missing_keys, unexpected_keys)
    return unet



def create_classifier_unet(
        image_size,
        classifier_width,
        class_number,
        classifier_depth,
        classifier_attention_resolutions,
        channel_mult,
        classifier_use_scale_shift_norm,
        classifier_resblock_updown,
        classifier_pool,
        pertrain_path=None,
    ):

    classifier=EncoderUNetModel(image_size=image_size,
                                in_channels=3,
                                model_channels=classifier_width,
                                out_channels=class_number,
                                num_res_blocks=classifier_depth,
                                attention_resolutions=classifier_attention_resolutions,
                                dropout=0,
                                channel_mult=channel_mult,
                                conv_resample=True,
                                num_heads=1,
                                num_heads_upsample=-1,
                                num_head_channels=64,
                                use_scale_shift_norm=classifier_use_scale_shift_norm,
                                resblock_updown=classifier_resblock_updown,
                                pool=classifier_pool,
                                )

    if pertrain_path is not None:
        missing_keys, unexpected_keys= load_pertrain_weights(pertrain_path,classifier)
        # print(missing_keys, unexpected_keys)

    return classifier




import numpy as np
import math


def lerp_np(x,y,w):
    fin_out = (y-x)*w + x
    return fin_out


def rand_perlin_2d_np(shape, res, fade=lambda t: 6 * t ** 5 - 15 * t ** 4 + 10 * t ** 3):
    delta = (res[0] / shape[0], res[1] / shape[1])
    d = (shape[0] // res[0], shape[1] // res[1])
    grid = np.mgrid[0:res[0]:delta[0], 0:res[1]:delta[1]].transpose(1, 2, 0) % 1

    angles = 2 * math.pi * np.random.rand(res[0] + 1, res[1] + 1)
    gradients = np.stack((np.cos(angles), np.sin(angles)), axis=-1)
    tt = np.repeat(np.repeat(gradients,d[0],axis=0),d[1],axis=1)

    tile_grads = lambda slice1, slice2: np.repeat(np.repeat(gradients[slice1[0]:slice1[1], slice2[0]:slice2[1]],d[0],axis=0),d[1],axis=1)
    dot = lambda grad, shift: (
                np.stack((grid[:shape[0], :shape[1], 0] + shift[0], grid[:shape[0], :shape[1], 1] + shift[1]),
                            axis=-1) * grad[:shape[0], :shape[1]]).sum(axis=-1)

    n00 = dot(tile_grads([0, -1], [0, -1]), [0, 0])
    n10 = dot(tile_grads([1, None], [0, -1]), [-1, 0])
    n01 = dot(tile_grads([0, -1], [1, None]), [0, -1])
    n11 = dot(tile_grads([1, None], [1, None]), [-1, -1])
    t = fade(grid[:shape[0], :shape[1]])
    return math.sqrt(2) * lerp_np(lerp_np(n00, n10, t[..., 0]), lerp_np(n01, n11, t[..., 0]), t[..., 1])


from torch import nn
from abc import abstractmethod
import torch
import torch.nn.functional as F
import math



def timestep_embedding(timesteps, dim, max_period=10000):
    """
    Create sinusoidal timestep embeddings.

    :param timesteps: a 1-D Tensor of N indices, one per batch element.
                      These may be fractional.
    :param dim: the dimension of the output.
    :param max_period: controls the minimum frequency of the embeddings.
    :return: an [N x dim] Tensor of positional embeddings.
    """
    half = dim // 2
    freqs = torch.exp(
        -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
    ).to(device=timesteps.device)
    args = timesteps[:, None].float() * freqs[None]
    embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
    if dim % 2:
        embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
    return embedding



class SiLU(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


class GroupNorm32(nn.GroupNorm):
    def forward(self, x):
        return super().forward(x.float()).type(x.dtype)



def normalization(channels):
    """
    Make a standard normalization layer.

    :param channels: number of input channels.
    :return: an nn.Module for normalization.
    """
    return GroupNorm32(32, channels)


def zero_module(module):
    """
    Zero out the parameters of a module and return it.
    """
    for p in module.parameters():
        p.detach().zero_()
    return module


def convert_module_to_f16(l):
    """
    Convert primitive modules to float16.
    """
    if isinstance(l, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        l.weight.data = l.weight.data.half()
        if l.bias is not None:
            l.bias.data = l.bias.data.half()


def convert_module_to_f32(l):
    """
    Convert primitive modules to float32, undoing convert_module_to_f16().
    """
    if isinstance(l, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        l.weight.data = l.weight.data.float()
        if l.bias is not None:
            l.bias.data = l.bias.data.float()


class TimestepBlock(nn.Module):
    """
    Any module where forward() takes timestep embeddings as a second argument.
    """

    @abstractmethod
    def forward(self, x, emb):
        """
        Apply the module to `x` given `emb` timestep embeddings.
        """


class TimestepEmbedSequential(nn.Sequential, TimestepBlock):
    """
    A sequential module that passes timestep embeddings to the children that
    support it as an extra input.
    """

    def forward(self, x, emb):
        for layer in self:
            if isinstance(layer, TimestepBlock):
                x = layer(x, emb)
            else:
                x = layer(x)
        return x


class Upsample(nn.Module):
    """
    An upsampling layer with an optional convolution.

    :param channels: channels in the inputs and outputs.
    :param use_conv: a bool determining if a convolution is applied.
    :param dims: determines if the signal is 1D, 2D, or 3D. If 3D, then
                 upsampling occurs in the inner-two dimensions.
    """

    def __init__(self, channels, use_conv, out_channels=None):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        if use_conv:
            self.conv = nn.Conv2d(self.channels, self.out_channels, 3, padding=1)


    def forward(self, x):
        assert x.shape[1] == self.channels
        x = F.interpolate(x, scale_factor=2, mode="nearest")
        if self.use_conv:
            x = self.conv(x)
        return x


class Downsample(nn.Module):
    """
    A downsampling layer with an optional convolution.

    :param channels: channels in the inputs and outputs.
    :param use_conv: a bool determining if a convolution is applied.
    :param dims: determines if the signal is 1D, 2D, or 3D. If 3D, then
                 downsampling occurs in the inner-two dimensions.
    """

    def __init__(self, channels, use_conv, out_channels=None):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        stride = 2
        if use_conv:
            self.op = nn.Conv2d(
                 self.channels, self.out_channels, 3, stride=stride, padding=1
            )
        else:
            assert self.channels == self.out_channels
            self.op = nn.AvgPool2d(kernel_size=stride, stride=stride)

    def forward(self, x):
        assert x.shape[1] == self.channels
        return self.op(x)


class QKVAttention(nn.Module):
    """
    A module which performs QKV attention and splits in a different order.
    """

    def __init__(self, n_heads):
        super().__init__()
        self.n_heads = n_heads

    def forward(self, qkv):
        """
        Apply QKV attention.

        :param qkv: an [N x (3 * H * C) x T] tensor of Qs, Ks, and Vs.
        :return: an [N x (H * C) x T] tensor after attention.
        """
        bs, width, length = qkv.shape
        assert width % (3 * self.n_heads) == 0
        ch = width // (3 * self.n_heads)
        q, k, v = qkv.chunk(3, dim=1)
        scale = 1 / math.sqrt(math.sqrt(ch))
        weight = torch.einsum(
            "bct,bcs->bts",
            (q * scale).view(bs * self.n_heads, ch, length),
            (k * scale).view(bs * self.n_heads, ch, length),
        )  # More stable with f16 than dividing afterwards
        weight = torch.softmax(weight.float(), dim=-1).type(weight.dtype)
        a = torch.einsum("bts,bcs->bct", weight, v.reshape(bs * self.n_heads, ch, length))
        return a.reshape(bs, -1, length)



class QKVAttentionLegacy(nn.Module):
    """
    A module which performs QKV attention. Matches legacy QKVAttention + input/ouput heads shaping
    """

    def __init__(self, n_heads):
        super().__init__()
        self.n_heads = n_heads

    def forward(self, qkv):
        """
        Apply QKV attention.

        :param qkv: an [N x (H * 3 * C) x T] tensor of Qs, Ks, and Vs.
        :return: an [N x (H * C) x T] tensor after attention.
        """
        bs, width, length = qkv.shape
        assert width % (3 * self.n_heads) == 0
        ch = width // (3 * self.n_heads)
        q, k, v = qkv.reshape(bs * self.n_heads, ch * 3, length).split(ch, dim=1)
        scale = 1 / math.sqrt(math.sqrt(ch))
        weight = torch.einsum(
            "bct,bcs->bts", q * scale, k * scale
        )  # More stable with f16 than dividing afterwards
        weight = torch.softmax(weight.float(), dim=-1).type(weight.dtype)
        a = torch.einsum("bts,bcs->bct", weight, v)
        return a.reshape(bs, -1, length)




class AttentionPool2d(nn.Module):
    """
    Adapted from CLIP: https://github.com/openai/CLIP/blob/main/clip/model.py
    """

    def __init__(
        self,
        spacial_dim: int,
        embed_dim: int,
        num_heads_channels: int,
        output_dim: int = None,
    ):
        super().__init__()
        self.positional_embedding = nn.Parameter(
            torch.randn(embed_dim, spacial_dim ** 2 + 1) / embed_dim ** 0.5
        )
        self.qkv_proj = nn.Conv1d( embed_dim, 3 * embed_dim, 1)
        self.c_proj = nn.Conv1d( embed_dim, output_dim or embed_dim, 1)
        self.num_heads = embed_dim // num_heads_channels
        self.attention = QKVAttention(self.num_heads)


    def forward(self, x):
        b, c, *_spatial = x.shape
        x = x.reshape(b, c, -1)  # NC(HW)
        x = torch.cat([x.mean(dim=-1, keepdim=True), x], dim=-1)  # NC(HW+1)
        x = x + self.positional_embedding[None, :, :].to(x.dtype)  # NC(HW+1)
        x = self.qkv_proj(x)
        x = self.attention(x)
        x = self.c_proj(x)
        return x[:, :, 0]




class ResBlock(TimestepBlock):
    """
    A residual block that can optionally change the number of channels.

    :param channels: the number of input channels.
    :param emb_channels: the number of timestep embedding channels.
    :param dropout: the rate of dropout.
    :param out_channels: if specified, the number of out channels.
    :param use_conv: if True and out_channels is specified, use a spatial
        convolution instead of a smaller 1x1 convolution to change the
        channels in the skip connection.
    :param dims: determines if the signal is 1D, 2D, or 3D.
    :param use_checkpoint: if True, use gradient checkpointing on this module.
    :param up: if True, use this block for upsampling.
    :param down: if True, use this block for downsampling.
    """

    def __init__(
        self,
        channels,
        emb_channels,
        dropout,
        out_channels=None,
        use_conv=False,
        use_scale_shift_norm=False,
        up=False,
        down=False,
    ):
        super().__init__()
        self.channels = channels
        self.emb_channels = emb_channels
        self.dropout = dropout
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.use_scale_shift_norm = use_scale_shift_norm

        self.in_layers = nn.Sequential(
            normalization(channels),
            nn.SiLU(),
            nn.Conv2d(channels, self.out_channels, 3, padding=1),
        )

        self.updown = up or down

        if up:
            self.h_upd = Upsample(channels, False)
            self.x_upd = Upsample(channels, False)
        elif down:
            self.h_upd = Downsample(channels, False)
            self.x_upd = Downsample(channels, False)
        else:
            self.h_upd = self.x_upd = nn.Identity()

        self.emb_layers = nn.Sequential(
            nn.SiLU(),
            nn.Linear(
                emb_channels,
                2 * self.out_channels if use_scale_shift_norm else self.out_channels,
            ),
        )
        self.out_layers = nn.Sequential(
            normalization(self.out_channels),
            nn.SiLU(),
            nn.Dropout(p=dropout),
            zero_module(
                nn.Conv2d(self.out_channels, self.out_channels, 3, padding=1)
            ),
        )

        if self.out_channels == channels:
            self.skip_connection = nn.Identity()
        elif use_conv:
            self.skip_connection = nn.Conv2d(
                 channels, self.out_channels, 3, padding=1
            )
        else:
            self.skip_connection = nn.Conv2d( channels, self.out_channels, 1)



    def forward(self, x, emb):
        if self.updown:
            in_rest, in_conv = self.in_layers[:-1], self.in_layers[-1]
            h = in_rest(x)
            h = self.h_upd(h)
            x = self.x_upd(x)
            h = in_conv(h)
        else:
            h = self.in_layers(x)
        emb_out = self.emb_layers(emb).type(h.dtype)
        while len(emb_out.shape) < len(h.shape):
            emb_out = emb_out[..., None]
        if self.use_scale_shift_norm:
            out_norm, out_rest = self.out_layers[0], self.out_layers[1:]
            scale, shift = torch.chunk(emb_out, 2, dim=1)
            h = out_norm(h) * (1 + scale) + shift
            h = out_rest(h)
        else:
            h = h + emb_out
            h = self.out_layers(h)
        return self.skip_connection(x) + h



class AttentionBlock(nn.Module):
    """
    An attention block that allows spatial positions to attend to each other.

    Originally ported from here, but adapted to the N-d case.
    https://github.com/hojonathanho/diffusion/blob/1e0dceb3b3495bbe19116a5e1b3596cd0706c543/diffusion_tf/models/unet.py#L66.
    """

    def __init__(
        self,
        channels,
        num_heads=1,
        num_head_channels=-1,
    ):
        super().__init__()
        self.channels = channels
        if num_head_channels == -1:
            self.num_heads = num_heads
        else:
            assert (
                channels % num_head_channels == 0
            ), f"q,k,v channels {channels} is not divisible by num_head_channels {num_head_channels}"
            self.num_heads = channels // num_head_channels
        self.norm = normalization(channels)
        self.qkv = nn.Conv1d(channels, channels * 3, 1)

        self.attention = QKVAttentionLegacy(self.num_heads)
        self.proj_out = zero_module(nn.Conv1d(channels, channels, 1))


    def forward(self, x):
        b, c, *spatial = x.shape
        x = x.reshape(b, c, -1)
        qkv = self.qkv(self.norm(x))
        h = self.attention(qkv)
        h = self.proj_out(h)
        return (x + h).reshape(b, c, *spatial)



"""Anomaly metrics."""
import numpy as np
import warnings
import torch
import torch.nn.functional as F
from torchmetrics.classification import BinaryAveragePrecision
from torchmetrics.utilities.compute import auc
from torchmetrics.classification.roc import BinaryROC
from torchmetrics.classification import BinaryPrecisionRecallCurve as _BinaryPrecisionRecallCurve
from torch import Tensor
from torchmetrics.functional.classification.precision_recall_curve import (
    _adjust_threshold_arg,
    _binary_precision_recall_curve_update,
)
from .pro import AUPRO
warnings.filterwarnings('ignore')

class BinaryPrecisionRecallCurve(_BinaryPrecisionRecallCurve):
    """Binary precision-recall curve with without threshold prediction normalization."""

    @staticmethod
    def _binary_precision_recall_curve_format(
        preds,
        target,
        thresholds = None,
        ignore_index = None,
    ) :
        """Similar to torchmetrics' ``_binary_precision_recall_curve_format`` except it does not apply sigmoid."""
        preds = preds.flatten()
        target = target.flatten()
        if ignore_index is not None:
            idx = target != ignore_index
            preds = preds[idx]
            target = target[idx]

        thresholds = _adjust_threshold_arg(thresholds, preds.device)
        return preds, target, thresholds

    def update(self, preds, target) -> None:
        """Update metric state with new predictions and targets.

        Unlike the base class, this accepts raw predictions and targets.

        Args:
            preds (Tensor): Predicted probabilities
            target (Tensor): Ground truth labels
        """
        preds, target, _ = BinaryPrecisionRecallCurve._binary_precision_recall_curve_format(
            preds,
            target,
            self.thresholds,
            self.ignore_index,
        )
        state = _binary_precision_recall_curve_update(preds, target, self.thresholds)
        if isinstance(state, Tensor):
            self.confmat += state
        else:
            self.preds.append(state[0])
            self.target.append(state[1])


def compute_imagewise_metrics_gpu(
    prediction_scores, gt_labels, device = None, **kwargs
):
    """
       Computes image-wise metrics (Image Auroc) on GPU.

       Args:
           prediction_scores: [np.array or list] [N] Assignment weights
                                       per image. Higher indicates higher
                                       probability of being an anomaly.
           gt_labels: [np.array or list] [N] Binary labels - 1
                                       if image is an anomaly, 0 if not.
           device: torch.device
       """
    if device is None:
        device = torch.device("cuda")
    prediction_scores = torch.tensor(prediction_scores).to(device)
    gt_labels = torch.tensor(gt_labels).to(device).long()

    fpr, tpr, thresholds = BinaryROC(thresholds=None)(
         prediction_scores, gt_labels
    )
    auroc = auc(fpr, tpr, reorder=True)
    maxindex = (tpr - fpr).tolist().index(max(tpr - fpr))
    threshold = thresholds[maxindex]

    return {"image_auroc": auroc.item(), "image_threshold": threshold.item()}


def compute_pixelwise_metrics_gpu(prediction_masks, gt_masks, device =None, **kwargs):
    """
     Computes pixel-wise metrics (Pixel-AUROC, AP, F1) on GPU

     Args:
         prediction_masks: [list of np.arrays or np.array] [NxHxW] Contains
                                 generated segmentation masks.
         gt_masks: [list of np.arrays or np.array] [NxHxW] Contains
                             predefined ground truth segmentation masks
         device: torch.device
    """

    if device is None:
        device = torch.device("cuda")

    if isinstance(prediction_masks, list):
        prediction_masks = np.stack(prediction_masks)

    if isinstance(gt_masks, list):
        gt_masks = np.stack(gt_masks)

    prediction_masks = torch.tensor(prediction_masks).to(device)
    gt_masks = torch.tensor(gt_masks).to(device).long()

    flat_anomaly_segmentations = prediction_masks.flatten()
    flat_ground_truth_masks = gt_masks.flatten()

    fpr, tpr, _ = BinaryROC(thresholds=None)(
        flat_anomaly_segmentations, flat_ground_truth_masks
    )

    auroc = auc(fpr, tpr, reorder=True)

    precision, recall, thresholds = BinaryPrecisionRecallCurve()(flat_anomaly_segmentations,flat_ground_truth_masks)
    f1_score = (2 * precision * recall) / (precision + recall + 1e-10)
    seg_threshold = thresholds[torch.argmax(f1_score)]
    f1 = torch.max(f1_score)

    ap = BinaryAveragePrecision(thresholds=None)(flat_anomaly_segmentations, flat_ground_truth_masks)

    return {
        "pixel_auroc": auroc.item(),
        "pixel_ap": ap.item(),
        "pixel_f1": f1.item(),
        "seg_threshold": seg_threshold.item()
    }


def compute_pro_gpu(prediction_masks, gt_masks, device = None, **kwargs):
    """Compute the area under the curve of per-region overlaping (PRO) and 0 to 0.3 FPR on GPU
       Args:
           prediction_masks (ndarray): All anomaly maps in test. amaps.shape -> (num_test_data, h, w)
           gt_masks (ndarray): All binary masks in test. masks.shape -> (num_test_data, h, w)
           device: torch.device
       """
    if device is None:
        device = torch.device("cuda")
    prediction_masks = torch.tensor(prediction_masks).to(device)
    gt_masks = torch.tensor(gt_masks).to(device).long()
    aupro = AUPRO(fpr_limit=0.3, num_thresholds=200)(prediction_masks, gt_masks)
    return {"pixel_pro": aupro.item()}



def compute_pixelwise_metrics_gpu_resize(prediction_masks, gt_masks, resize, device = None, **kwargs):
    if device is None:
        device = torch.device("cuda")
    prediction_masks_resize = F.interpolate(torch.tensor(prediction_masks[:,None,:,:]),
                                         (resize, resize), mode = 'bilinear').to(device).squeeze(1)
    gt_masks_resize = F.interpolate(torch.tensor(gt_masks[:, None, :, :]),
                                         (resize, resize), mode='nearest').to(device).squeeze(1)
    return compute_pixelwise_metrics_gpu(prediction_masks_resize, gt_masks_resize, device)


def compute_pro_gpu_resize(prediction_masks, gt_masks, resize,device = None, **kwargs):
    if device is None:
        device = torch.device("cuda")
    prediction_masks_resize = F.interpolate(torch.tensor(prediction_masks[:, None, :, :]),
                                            (resize, resize), mode='bilinear').to(device).squeeze(1)
    gt_masks_resize = F.interpolate(torch.tensor(gt_masks[:, None, :, :]),
                                    (resize, resize), mode='nearest').to(device).squeeze(1)
    return compute_pro_gpu(prediction_masks_resize, gt_masks_resize, device)



import torch
import torch.nn as nn
from torch.nn import functional as F
import geomloss
from typing import List

class ProjLayer(nn.Module):
    '''
    inputs: features of encoder block
    outputs: projected features
    '''
    def __init__(self, in_c, out_c):
        super(ProjLayer, self).__init__()
        self.proj = nn.Sequential(nn.Conv2d(in_c, in_c//2, kernel_size=3, stride=1, padding=1),
                                  nn.InstanceNorm2d(in_c//2),
                                  torch.nn.LeakyReLU(),
                                  nn.Conv2d(in_c//2, in_c//4, kernel_size=3, stride=1, padding=1),
                                  nn.InstanceNorm2d(in_c//4),
                                  torch.nn.LeakyReLU(),
                                  nn.Conv2d(in_c//4, in_c//2, kernel_size=3, stride=1, padding=1),
                                  nn.InstanceNorm2d(in_c//2),
                                  torch.nn.LeakyReLU(),
                                  nn.Conv2d(in_c//2, out_c, kernel_size=3, stride=1, padding=1),
                                  nn.InstanceNorm2d(out_c),
                                  torch.nn.LeakyReLU(),
                                  )
    def forward(self, x):
        return self.proj(x)


class MultiProjectionLayer(nn.Module):
    def __init__(self, base, layers_to_extract_from):
        super(MultiProjectionLayer, self).__init__()
        self.projs = []
        if 'layer1' in layers_to_extract_from:
            self.projs.append(ProjLayer(base * 4, base * 4))
        if 'layer2' in layers_to_extract_from:
            self.projs.append(ProjLayer(base * 8, base * 8))
        if 'layer3' in layers_to_extract_from:
            self.projs.append(ProjLayer(base * 16, base * 16))

        for index, proj in enumerate(self.projs):
            self.add_module(f"proj_{index}", proj)

    def forward(self, features, features_noise: List = False):
        if features_noise is not False:
            return ( [ proj(feature) for feature, proj in zip(features_noise, self.projs)],
                    [ proj(feature) for feature, proj in zip(features, self.projs)])
        else:
            return [ proj(feature) for feature, proj in zip(features, self.projs)]


def loss_fucntion(a, b):
    cos_loss = torch.nn.CosineSimilarity()
    loss = 0
    for item in range(len(a)):
        loss += torch.mean(1-cos_loss(a[item].view(a[item].shape[0],-1),
                                      b[item].view(b[item].shape[0],-1)))
    return loss


def loss_concat(a, b):
    mse_loss = torch.nn.MSELoss()
    cos_loss = torch.nn.CosineSimilarity()
    loss = 0
    a_map = []
    b_map = []
    size = a[0].shape[-1]
    for item in range(len(a)):
        a_map.append(F.interpolate(a[item], size=size, mode='bilinear', align_corners=True))
        b_map.append(F.interpolate(b[item], size=size, mode='bilinear', align_corners=True))
    a_map = torch.cat(a_map,1)
    b_map = torch.cat(b_map,1)
    loss += torch.mean(1-cos_loss(a_map,b_map))
    return loss


class CosineReconstruct(nn.Module):
    def __init__(self):
        super(CosineReconstruct, self).__init__()
    def forward(self, x, y):
        return torch.mean(1 - torch.nn.CosineSimilarity()(x, y))


class Revisit_RDLoss(nn.Module):
    """
    receive multiple inputs feature
    return multi-task loss:  SSOT loss, Reconstruct Loss, Contrast Loss
    """
    def __init__(self, consistent_shuffle = True):
        super(Revisit_RDLoss, self).__init__()
        self.sinkhorn = geomloss.SamplesLoss(loss='sinkhorn', p=2, blur=0.05, \
                              reach=None, diameter=10000000, scaling=0.95, \
                                truncate=10, cost=None, kernel=None, cluster_scale=None, \
                                  debias=True, potentials=False, verbose=False, backend='tensorized')

        self.reconstruct = CosineReconstruct()       
        self.contrast = torch.nn.CosineEmbeddingLoss(margin = 0.5)

    def forward(self, noised_features, projected_noised_features, projected_normal_features):
        """
        noised_feature : output of encoder at each_blocks : [noised_feature_block1, noised_feature_block2, noised_feature_block3]
        projected_noised_feature: list of the projection layer's output on noised_features, projected_noised_feature = projection(noised_feature)
        projected_normal_feature: list of the projection layer's output on normal_features, projected_normal_feature = projection(normal_feature)
        """
        current_batchsize = projected_normal_features[0].shape[0]

        target = -torch.ones(current_batchsize).to('cuda')

        shuffle_index = torch.randperm(current_batchsize)

        shuffle_features  = [ feature[shuffle_index] for feature in projected_normal_features]

        loss_ssot = torch.sum(torch.stack([self.sinkhorn(torch.softmax(normal_proj.view(normal_proj.shape[0], -1), -1), torch.softmax(shuffle_feature.view(shuffle_feature.shape[0], -1), -1))
                               for normal_proj, shuffle_feature in zip(projected_normal_features, shuffle_features) ]))

        loss_reconstruct = torch.sum(torch.stack([self.reconstruct(abnormal_proj, normal_proj) for abnormal_proj, normal_proj in zip(projected_noised_features, projected_normal_features )]))

        loss_contrast = torch.sum(torch.stack([ self.contrast(noised_feature.view(noised_feature.shape[0], -1), normal_proj.view(normal_proj.shape[0], -1), target = target) for noised_feature, normal_proj in zip(noised_features, projected_normal_features )    ]))

        return (loss_ssot + 0.01 * loss_reconstruct + 0.1 * loss_contrast) / 1.11



import timm  # noqa
import torchvision.models as models  # noqa
import torch
import copy

class ForwardHook:
    def __init__(self, hook_dict, layer_name: str, last_layer_to_extract: str):
        self.hook_dict = hook_dict
        self.layer_name = layer_name
        self.raise_exception_to_break = copy.deepcopy(
            layer_name == last_layer_to_extract
        )

    def __call__(self, module, input, output):
        self.hook_dict[self.layer_name] = output
        if self.raise_exception_to_break:
            raise LastLayerToExtractReachedException()
        return None


class LastLayerToExtractReachedException(Exception):
    pass


class NetworkFeatureAggregator(torch.nn.Module):
    """Efficient extraction of network features."""

    def __init__(self, backbone, layers_to_extract_from, device):
        super(NetworkFeatureAggregator, self).__init__()
        """Extraction of network features.

        Runs a network only to the last layer of the list of layers where
        network features should be extracted from.

        Args:
            backbone: torchvision.model
            layers_to_extract_from: [list of str]
        """

        self.layers_to_extract_from = layers_to_extract_from
        self.backbone = backbone
        self.device = device

        if not hasattr(backbone, "hook_handles"):
            self.backbone.hook_handles = []

        for handle in self.backbone.hook_handles:
            handle.remove()

        self.outputs = {}

        for extract_layer in layers_to_extract_from:

            forward_hook = ForwardHook(
                self.outputs, extract_layer, layers_to_extract_from[-1]
            )
            if "." in extract_layer:
                extract_block, extract_idx = extract_layer.split(".")
                network_layer = backbone.__dict__["_modules"][extract_block]
                if extract_idx.isnumeric():
                    extract_idx = int(extract_idx)
                    network_layer = network_layer[extract_idx]
                else:
                    network_layer = network_layer.__dict__["_modules"][extract_idx]
            else:
                # print(list(backbone.__dict__["_modules"].keys()))
                network_layer = backbone.__dict__["_modules"][extract_layer]

            if isinstance(network_layer, torch.nn.Sequential):
                self.backbone.hook_handles.append(
                    network_layer[-1].register_forward_hook(forward_hook)
                )
            else:
                self.backbone.hook_handles.append(
                    network_layer.register_forward_hook(forward_hook)
                )
        self.to(self.device)


    def forward(self, images):
        self.outputs.clear()
        with torch.no_grad():
            # The backbone will throw an Exception once it reached the last
            # layer to compute features from. Computation will stop there.
            try:
                _ = self.backbone(images)
            except LastLayerToExtractReachedException:
                pass
        return self.outputs


_BACKBONES = {
    "alexnet": "models.alexnet(pretrained=True)",
    "bninception": 'pretrainedmodels.__dict__["bninception"]'
    '(pretrained="imagenet", num_classes=1000)',
    "resnet50": "models.resnet50(pretrained=True)",
    "resnet18": "models.resnet18(pretrained=True)",
    "resnet101": "models.resnet101(pretrained=True)",
    "resnext101": "models.resnext101_32x8d(pretrained=True)",
    "resnet200": 'timm.create_model("resnet200", pretrained=True)',
    "resnest50": 'timm.create_model("resnest50d_4s2x40d", pretrained=True)',
    "resnetv2_50_bit": 'timm.create_model("resnetv2_50x3_bitm", pretrained=True)',
    "resnetv2_50_21k": 'timm.create_model("resnetv2_50x3_bitm_in21k", pretrained=True)',
    "resnetv2_101_bit": 'timm.create_model("resnetv2_101x3_bitm", pretrained=True)',
    "resnetv2_101_21k": 'timm.create_model("resnetv2_101x3_bitm_in21k", pretrained=True)',
    "resnetv2_152_bit": 'timm.create_model("resnetv2_152x4_bitm", pretrained=True)',
    "resnetv2_152_21k": 'timm.create_model("resnetv2_152x4_bitm_in21k", pretrained=True)',
    "resnetv2_152_384": 'timm.create_model("resnetv2_152x2_bit_teacher_384", pretrained=True)',
    "resnetv2_101": 'timm.create_model("resnetv2_101", pretrained=True)',
    "vgg11": "models.vgg11(pretrained=True)",
    "vgg19": "models.vgg19(pretrained=True)",
    "vgg19_bn": "models.vgg19_bn(pretrained=True)",
    "wideresnet50": "models.wide_resnet50_2(pretrained=True)",
    "wideresnet101": "models.wide_resnet101_2(pretrained=True)",
    "mnasnet_100": 'timm.create_model("mnasnet_100", pretrained=True)',
    "mnasnet_a1": 'timm.create_model("mnasnet_a1", pretrained=True)',
    "mnasnet_b1": 'timm.create_model("mnasnet_b1", pretrained=True)',
    "densenet121": 'timm.create_model("densenet121", pretrained=True)',
    "densenet201": 'timm.create_model("densenet201", pretrained=True)',
    "inception_v4": 'timm.create_model("inception_v4", pretrained=True)',
    "vit_small": 'timm.create_model("vit_small_patch16_224", pretrained=True)',
    "vit_base": 'timm.create_model("vit_base_patch16_224", pretrained=True)',
    "vit_large": 'timm.create_model("vit_large_patch16_224", pretrained=True)',
    "vit_r50": 'timm.create_model("vit_large_r50_s32_224", pretrained=True)',
    "vit_base_patch8_224": 'timm.create_model("vit_base_patch8_224", pretrained=True)',
    "vit_base_patch16_224": 'timm.create_model("vit_base_patch16_224", pretrained=True)',
    "vit_base_patch32_224": 'timm.create_model("vit_base_patch32_224", pretrained=True)',
    "vit_base_patch16_384": 'timm.create_model("vit_base_patch16_384", pretrained=True)',
    "vit_deit_base": 'timm.create_model("deit_base_patch16_224", pretrained=True)',
    "vit_deit_distilled": 'timm.create_model("deit_base_distilled_patch16_224", pretrained=True)',
    "vit_swin_base": 'timm.create_model("swin_base_patch4_window7_224", pretrained=True)',
    "vit_swin_large": 'timm.create_model("swin_large_patch4_window7_224", pretrained=True)',
    "efficientnet_b7": 'timm.create_model("tf_efficientnet_b7", pretrained=True)',
    "efficientnet_b5": 'timm.create_model("tf_efficientnet_b5", pretrained=True)',
    "efficientnet_b3": 'timm.create_model("tf_efficientnet_b3", pretrained=True)',
    "efficientnet_b1": 'timm.create_model("tf_efficientnet_b1", pretrained=True)',
    "efficientnetv2_m": 'timm.create_model("tf_efficientnetv2_m", pretrained=True)',
    "efficientnetv2_l": 'timm.create_model("tf_efficientnetv2_l", pretrained=True)',
    "efficientnet_b3a": 'timm.create_model("efficientnet_b3a", pretrained=True)',
    "coc_medium": 'timm.create_model("coc_medium", pretrained=True)',
}


def load(name):
    return eval(_BACKBONES[name])




import glob
import logging
import os
import numpy as np
import tabulate
import torch
import torch.nn.functional as F
from sklearn import metrics
from skimage import measure


class Report:
    def __init__(self, heads=None):
        if heads:
            self.heads = list(map(str, heads))
        else:
            self.heads = ()
        self.records = []

    def add_one_record(self, record):
        if self.heads:
            if len(record) != len(self.heads):
                raise ValueError(
                    f"Record's length ({len(record)}) should be equal to head's length ({len(self.heads)})."
                )
        self.records.append(record)

    def __str__(self):
        return tabulate.tabulate(
            self.records,
            self.heads,
            tablefmt="pipe",
            numalign="center",
            stralign="center",
        )



class EvalDataMeta:
    def __init__(self, preds, masks):
        self.preds = preds  # N x H x W
        self.masks = masks  # N x H x W


class EvalImage:
    def __init__(self, data_meta, **kwargs):
        self.preds = self.encode_pred(data_meta.preds, **kwargs)
        self.masks = self.encode_mask(data_meta.masks)
        self.preds_good = sorted(self.preds[self.masks == 0], reverse=True)
        self.preds_defe = sorted(self.preds[self.masks == 1], reverse=True)
        self.num_good = len(self.preds_good)
        self.num_defe = len(self.preds_defe)

    @staticmethod
    def encode_pred(preds):
        raise NotImplementedError

    def encode_mask(self, masks):
        N, _, _ = masks.shape
        masks = (masks.reshape(N, -1).sum(axis=1) != 0).astype(np.int)  # (N, )
        return masks

    def eval_auc(self):
        fpr, tpr, thresholds = metrics.roc_curve(self.masks, self.preds, pos_label=1)
        auc = metrics.auc(fpr, tpr)
        if auc < 0.5:
            auc = 1 - auc
        return auc


class EvalImageMax(EvalImage):
    @staticmethod
    def encode_pred(preds, avgpool_size):
        N, _, _ = preds.shape
        preds = torch.tensor(preds[:, None, ...]).cuda()  # N x 1 x H x W
        preds = (
            F.avg_pool2d(preds, avgpool_size, stride=1).cpu().numpy()
        )  # N x 1 x H x W
        return preds.reshape(N, -1).max(axis=1)  # (N, )



class EvalPixelAUC:
    def __init__(self, data_meta):
        self.preds = np.concatenate(
            [pred.flatten() for pred in data_meta.preds], axis=0
        )
        self.masks = np.concatenate(
            [mask.flatten() for mask in data_meta.masks], axis=0
        )
        self.masks[self.masks > 0] = 1

    def eval_auc(self):
        fpr, tpr, thresholds = metrics.roc_curve(self.masks, self.preds, pos_label=1)
        auc = metrics.auc(fpr, tpr)
        if auc < 0.5:
            auc = 1 - auc
        return auc


class EvalPixelPro:
    def __init__(self, data_meta):
        self.preds = np.concatenate(
            [np.expand_dims(pred,axis=0) for pred in data_meta.preds], axis=0
        )
        self.masks = np.concatenate(
            [np.expand_dims(mask,axis=0)  for mask in data_meta.masks], axis=0
        )
        self.masks[self.masks > 0] = 1
        self.masks=self.masks.astype(np.bool)

    def rescale(self,x):
        return (x - x.min()) / (x.max() - x.min())

    def eval_auc(self):

        max_step = 1000
        expect_fpr = 0.3  # default 30%

        max_th = self.preds.max()
        min_th = self.preds.min()
        delta = (max_th - min_th) / max_step

        pros_mean = []
        pros_std = []

        threds = []
        fprs = []

        binary_score_maps = np.zeros_like(self.preds, dtype=np.bool)

        for step in range(max_step):
            thred = max_th - step * delta
            # segmentation
            binary_score_maps[self.preds <= thred] = 0
            binary_score_maps[self.preds > thred] = 1
            pro = []  # per region overlap
            # pro: find each connected gt region, compute the overlapped pixels between the gt region and predicted region
            # iou: for each image, compute the ratio, i.e. intersection/union between the gt and predicted binary map
            for i in range(len(binary_score_maps)):  # for i th image
                # pro (per region level)
                label_map = measure.label(self.masks[i], connectivity=2)
                props = measure.regionprops(label_map)
                for prop in props:
                    x_min, y_min, x_max, y_max = prop.bbox  # find the bounding box of an anomaly region
                    cropped_pred_label = binary_score_maps[i][x_min:x_max, y_min:y_max]
                    # cropped_mask = gt_mask[i][x_min:x_max, y_min:y_max]   # bug!
                    cropped_mask = prop.filled_image  # corrected!
                    intersection = np.logical_and(cropped_pred_label, cropped_mask).astype(np.float32).sum()
                    pro.append(intersection / prop.area)

            pros_mean.append(np.array(pro).mean())
            pros_std.append(np.array(pro).std())
            # fpr for pro-auc
            gt_masks_neg = ~self.masks
            fpr = np.logical_and(gt_masks_neg, binary_score_maps).sum() / gt_masks_neg.sum()
            fprs.append(fpr)
            threds.append(thred)
        # as array
        pros_mean = np.array(pros_mean)
        fprs = np.array(fprs)
        idx = fprs <= expect_fpr  # find the indexs of fprs that is less than expect_fpr (default 0.3)
        fprs_selected = fprs[idx]
        fprs_selected = self.rescale(fprs_selected)  # rescale fpr [0,0.3] -> [0, 1]
        pros_mean_selected = pros_mean[idx]
        seg_pro_auc = metrics.auc(fprs_selected, pros_mean_selected)
        return seg_pro_auc


eval_lookup_table = {
    "image": EvalImageMax,
    "pixel": EvalPixelAUC,
    "pro":EvalPixelPro,
}


def performances(class_name, preds, masks, config):
    ret_metrics = {}

    clsnames = [class_name]

    for clsname in clsnames:
        preds_cls = []
        masks_cls = []
        for pred, mask in zip( preds, masks):
            preds_cls.append(pred[None, ...])
            masks_cls.append(mask[None, ...])

        preds_cls = np.concatenate(np.asarray(preds_cls), axis=0)  # N x H x W
        masks_cls = np.concatenate(np.asarray(masks_cls), axis=0)  # N x H x W

        data_meta = EvalDataMeta(preds_cls, masks_cls)

        if config.get("auc", None):
            for metric in config.auc:
                evalname = metric["name"]
                kwargs = metric.get("kwargs", {})
                eval_method = eval_lookup_table[evalname](data_meta, **kwargs)
                auc = eval_method.eval_auc()
                ret_metrics["{}_{}_auc".format(clsname, evalname)] = auc


    if config.get("auc", None):
        for metric in config.auc:
            evalname = metric["name"]
            evalvalues = [
                ret_metrics["{}_{}_auc".format(clsname, evalname)]
                for clsname in clsnames
            ]
            mean_auc = np.mean(np.array(evalvalues))
            ret_metrics["{}_{}_auc".format("mean", evalname)] = mean_auc

    return ret_metrics



def log_metrics(ret_metrics, config,logger_name):
    logger = logging.getLogger(logger_name)
    clsnames = set([k.rsplit("_", 2)[0] for k in ret_metrics.keys()])
    clsnames = list(clsnames - set(["mean","best"]))
    clsnames.sort()
    clsnames = clsnames+ ['best']

    # auc
    if config.get("auc", None):
        auc_keys = [k for k in ret_metrics.keys() if "auc" in k]
        evalnames = list(set([k.rsplit("_", 2)[1] for k in auc_keys]))
        evalnames.sort()

        record = Report(["clsname"] + evalnames)

        for clsname in clsnames:
            clsvalues = [
                ret_metrics["{}_{}_auc".format(clsname, evalname)]
                for evalname in evalnames
            ]
            record.add_one_record([clsname] + clsvalues)

        logger.info(f"\n{record}")



import torch.multiprocessing as mp
import numpy as np
import torch.nn.functional as F
import torch
from sklearn.neighbors import NearestCentroid

from hiad.utils.device import embedding_in_device
from hiad.utils.split_and_gather import MultiResolutionIndex
from hiad.utils.base import split_list

def patch_embedding(samples,
                    indexes,
                    detector_class,
                    detector_config,
                    batch_size,
                    feature_resolution,
                    fusion_weights,
                    devices):

    num_device = len(devices)
    samples_in_device = split_list(samples, num_device)
    process_pool = mp.Pool(processes=num_device)
    embeddings = []

    for gpu_id, samples_ in zip(devices, samples_in_device):
        embedding = process_pool.apply_async(embedding_in_device, args=(gpu_id, detector_class,
                                                                        detector_config,
                                                                        samples_, indexes, batch_size,
                                                                        fusion_weights,
                                                                        feature_resolution))
        embeddings.append(embedding)

    process_pool.close()
    process_pool.join()
    patch_embeddings = []
    for embedding in embeddings:
        embedding = embedding.get()
        patch_embeddings.extend(embedding)
    return patch_embeddings


def update_paths_and_indexes(
                    tasks,
                    samples,
                    detector_class,
                    detector_config,
                    batch_size,
                    devices):

    if 'feature_resolution' in tasks[0]:
        feature_resolution = tasks[0]['feature_resolution']
        for task in tasks:
            assert feature_resolution == task['feature_resolution']
    else:
        feature_resolution = 1

    if 'fusion_weights' in tasks[0]:
        fusion_weights = tasks[0]['fusion_weights']
    else:
        fusion_weights = None

    indexes = []
    centers = []
    for task in tasks:
        centers.append(task['center'])
        for _, index in task['paths_and_indexes']:
            indexes.append(str(index))

    indexes = [MultiResolutionIndex.from_str(index) for index in set(indexes)]

    embedding_pairs = patch_embedding(samples,
                                    indexes,
                                    detector_class,
                                    detector_config,
                                    batch_size,
                                    feature_resolution,
                                    fusion_weights,
                                    devices)

    embeddings = [pair[-1].reshape(-1) for pair in embedding_pairs]
    paths_and_indexes = [pair[:-1] for pair in embedding_pairs]

    embeddings = np.stack(embeddings)
    embeddings = F.normalize(torch.tensor(embeddings), p=2, dim=1).numpy()

    clf = NearestCentroid()
    clf.centroids_ = np.stack(centers)
    clf.classes_ = np.arange(np.stack(centers).shape[0])
    labels = clf.predict(embeddings)

    for i, task in enumerate(tasks):
        task['paths_and_indexes'] = [ path_and_index  for label, path_and_index in  zip(labels, paths_and_indexes) if label == i]
    return tasks


import torch
from torch import optim as optim
from timm.optim.adafactor import Adafactor
from timm.optim.adahessian import Adahessian
from timm.optim.adamp import AdamP
from timm.optim.nadam import Nadam
from timm.optim.radam import RAdam
from timm.optim.rmsprop_tf import RMSpropTF
from timm.optim.sgdp import SGDP
from timm.optim.lookahead import Lookahead


def check_keywords_in_name(name, keywords=()):
	isin = False
	for keyword in keywords:
		if keyword in name:
			isin = True
	return isin


def add_weight_decay(model, weight_decay=1e-5, skip_list=(), skip_keywords=()):
	decay = []
	no_decay = []
	for name, param in model.named_parameters():
		if not param.requires_grad:
			continue  # frozen weights
		if len(param.shape) == 1 or name.endswith(".bias") or (name in skip_list) or check_keywords_in_name(name, skip_keywords):
			no_decay.append(param)
		else:
			decay.append(param)
	return [
		{'params': no_decay, 'weight_decay': 0.},
		{'params': decay, 'weight_decay': weight_decay}]


def get_optim(optim_kwargs, net, lr, betas=None, filter_bias_and_bn=True):
	kwargs = {k: v for k, v in optim_kwargs.items()}
	optim_split = kwargs.pop('name').lower().split('_')
	optim_name = optim_split[-1]
	optim_lookahead = optim_split[0] if len(optim_split) == 2 else None
	
	if kwargs.get('weight_decay', None) and filter_bias_and_bn:
		skip = {}
		skip_keywords = {}
		if hasattr(net, 'no_weight_decay'):
			skip = net.no_weight_decay()
		if hasattr(net, 'no_weight_decay_keywords'):
			skip_keywords = net.no_weight_decay_keywords()
		params = add_weight_decay(net, kwargs['weight_decay'], skip, skip_keywords)
		kwargs['weight_decay'] = 0.
	else:
		params = net.parameters()
		
	if kwargs.get('betas', None) and betas:
		kwargs['betas'] = betas
	
	optim_terms = {
		'sgd': optim.SGD,
		'adam': optim.Adam,
		'adamw': optim.AdamW,
		'adadelta': optim.Adadelta,
		'rmsprop': optim.RMSprop,
		'nadam': Nadam,
		'radam': RAdam,
		'adamp': AdamP,
		'sgdp': SGDP,
		'adafactor': Adafactor,
		'adahessian': Adahessian,
		'rmsproptf': RMSpropTF,
	}
	optimizer = optim_terms[optim_name](params, lr=lr, **kwargs)
	if optim_lookahead:
		optimizer = Lookahead(optimizer)
	return optimizer



import torch
import torch.nn as nn
import torch.nn.functional as F
from math import exp
from . import LOSS


@LOSS.register_module
class L1Loss(nn.Module):
    def __init__(self, lam=1):
        super(L1Loss, self).__init__()
        self.loss = nn.L1Loss(reduction='mean')
        self.lam = lam

    def forward(self, input1, input2):
        input1 = input1 if isinstance(input1, list) else [input1]
        input2 = input2 if isinstance(input2, list) else [input2]
        loss = 0
        for in1, in2 in zip(input1, input2):
            loss += self.loss(in1, in2) * self.lam
        return loss


@LOSS.register_module
class L2Loss(nn.Module):
    def __init__(self, lam=1):
        super(L2Loss, self).__init__()
        self.loss = nn.MSELoss(reduction='mean')
        self.lam = lam

    def forward(self, input1, input2):
        input1 = input1 if isinstance(input1, list) else [input1]
        input2 = input2 if isinstance(input2, list) else [input2]
        loss = 0
        for in1, in2 in zip(input1, input2):
            loss += self.loss(in1, in2) * self.lam
        return loss


@LOSS.register_module
class CosLoss(nn.Module):
    def __init__(self, avg=True, flat=True, lam=1):
        super(CosLoss, self).__init__()
        self.cos_sim = nn.CosineSimilarity()
        self.lam = lam
        self.avg = avg
        self.flat = flat

    def forward(self, input1, input2):
        input1 = input1 if isinstance(input1, list) else [input1]
        input2 = input2 if isinstance(input2, list) else [input2]
        loss = 0
        for in1, in2 in zip(input1, input2):
            if self.flat:
                loss += (1 - self.cos_sim(in1.contiguous().view(in1.shape[0], -1), in2.contiguous().view(in2.shape[0], -1))).mean() * self.lam
            else:
                loss += (1 - self.cos_sim(in1.contiguous(), in2.contiguous())).mean() * self.lam
        return loss / len(input1) if self.avg else loss


@LOSS.register_module
class KLLoss(nn.Module):
    def __init__(self, lam=1):
        super(KLLoss, self).__init__()
        self.loss = nn.KLDivLoss(reduction='mean')
        self.lam = lam

    def forward(self, input1, input2):
        # real, pred
        # teacher, student
        input1 = input1 if isinstance(input1, list) else [input1]
        input2 = input2 if isinstance(input2, list) else [input2]
        loss = 0
        for in1, in2 in zip(input1, input2):
            in1 = in1.permute(0, 2, 3, 1)
            in2 = in2.permute(0, 2, 3, 1)
            loss += self.loss(F.log_softmax(in2, dim=-1), F.softmax(in1, dim=-1)) * self.lam
        return loss


@LOSS.register_module
class LPIPSLoss(nn.Module):
    def __init__(self, lam=1):
        super(LPIPSLoss, self).__init__()
        self.loss = None
        self.lam = lam

    def forward(self, input1, input2):
        pass


@LOSS.register_module
class FocalLoss(nn.Module):
    """
    copy from: https://github.com/Hsuxu/Loss_ToolBox-PyTorch/blob/master/FocalLoss/FocalLoss.py
    This is a implementation of Focal Loss with smooth label cross entropy supported which is proposed in
    'Focal Loss for Dense Object Detection. (https://arxiv.org/abs/1708.02002)'
        Focal_Loss= -1*alpha*(1-pt)*log(pt)
    :param alpha: (tensor) 3D or 4D the scalar factor for this criterion
    :param gamma: (float,double) gamma > 0 reduces the relative loss for well-classified examples (p>0.5) putting more
                    focus on hard misclassified example
    :param smooth: (float,double) smooth value when cross entropy
    :param balance_index: (int) balance class index, should be specific when alpha is float
    :param size_average: (bool, optional) By default, the losses are averaged over each loss element in the batch.
    """

    def __init__(self, apply_nonlin=None, alpha=None, gamma=2, balance_index=0, smooth=1e-5, size_average=True, lam=1):
        super(FocalLoss, self).__init__()
        self.apply_nonlin = apply_nonlin
        self.alpha = alpha
        self.gamma = gamma
        self.balance_index = balance_index
        self.smooth = smooth
        self.size_average = size_average

        if self.smooth is not None:
            if self.smooth < 0 or self.smooth > 1.0:
                raise ValueError('smooth value should be in [0,1]')
        self.lam = lam

    def forward(self, logit, target):
        if self.apply_nonlin is not None:
            logit = self.apply_nonlin(logit)
        num_class = logit.shape[1]

        if logit.dim() > 2:
            # N,C,d1,d2 -> N,C,m (m=d1*d2*...)
            logit = logit.view(logit.size(0), logit.size(1), -1)
            logit = logit.permute(0, 2, 1).contiguous()
            logit = logit.view(-1, logit.size(-1))
        target = torch.squeeze(target, 1)
        target = target.view(-1, 1)
        alpha = self.alpha

        if alpha is None:
            alpha = torch.ones(num_class, 1)
        elif isinstance(alpha, (list, np.ndarray)):
            assert len(alpha) == num_class
            alpha = torch.FloatTensor(alpha).view(num_class, 1)
            alpha = alpha / alpha.sum()
        elif isinstance(alpha, float):
            alpha = torch.ones(num_class, 1)
            alpha = alpha * (1 - self.alpha)
            alpha[self.balance_index] = self.alpha

        else:
            raise TypeError('Not support alpha type')

        if alpha.device != logit.device:
            alpha = alpha.to(logit.device)

        idx = target.cpu().long()

        one_hot_key = torch.FloatTensor(target.size(0), num_class).zero_()
        one_hot_key = one_hot_key.scatter_(1, idx, 1)
        if one_hot_key.device != logit.device:
            one_hot_key = one_hot_key.to(logit.device)

        if self.smooth:
            one_hot_key = torch.clamp(
                one_hot_key, self.smooth / (num_class - 1), 1.0 - self.smooth)
        pt = (one_hot_key * logit).sum(1) + self.smooth
        logpt = pt.log()

        gamma = self.gamma

        alpha = alpha[idx]
        alpha = torch.squeeze(alpha)
        loss = -1 * alpha * torch.pow((1 - pt), gamma) * logpt

        if self.size_average:
            loss = loss.mean()
        loss *= self.lam
        return loss


def gaussian(window_size, sigma):
    gauss = torch.Tensor([exp(-(x - window_size//2)**2/float(2*sigma**2)) for x in range(window_size)])
    return gauss/gauss.sum()


def create_window(window_size, channel=1):
    _1D_window = gaussian(window_size, 1.5).unsqueeze(1)
    _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0)
    window = _2D_window.expand(channel, 1, window_size, window_size).contiguous()
    return window


def ssim(img1, img2, window_size=11, window=None, size_average=True, full=False, val_range=None):
    if val_range is None:
        if torch.max(img1) > 128:
            max_val = 255
        else:
            max_val = 1

        if torch.min(img1) < -0.5:
            min_val = -1
        else:
            min_val = 0
        l = max_val - min_val
    else:
        l = val_range

    padd = window_size//2
    (_, channel, height, width) = img1.size()
    if window is None:
        real_size = min(window_size, height, width)
        window = create_window(real_size, channel=channel).to(img1.device)

    mu1 = F.conv2d(img1, window, padding=padd, groups=channel)
    mu2 = F.conv2d(img2, window, padding=padd, groups=channel)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = F.conv2d(img1 * img1, window, padding=padd, groups=channel) - mu1_sq
    sigma2_sq = F.conv2d(img2 * img2, window, padding=padd, groups=channel) - mu2_sq
    sigma12 = F.conv2d(img1 * img2, window, padding=padd, groups=channel) - mu1_mu2

    c1 = (0.01 * l) ** 2
    c2 = (0.03 * l) ** 2

    v1 = 2.0 * sigma12 + c2
    v2 = sigma1_sq + sigma2_sq + c2
    cs = torch.mean(v1 / v2)  # contrast sensitivity

    ssim_map = ((2 * mu1_mu2 + c1) * v1) / ((mu1_sq + mu2_sq + c1) * v2)

    if size_average:
        ret = ssim_map.mean()
    else:
        ret = ssim_map.mean(1).mean(1).mean(1)

    if full:
        return ret, cs
    return ret, ssim_map


@LOSS.register_module
class SSIMLoss(torch.nn.Module):
    def __init__(self, window_size=11, size_average=True, val_range=None, lam=1):
        super(SSIMLoss, self).__init__()
        self.window_size = window_size
        self.size_average = size_average
        self.val_range = val_range

        # Assume 1 channel for SSIM
        self.channel = 1
        self.window = create_window(window_size).cuda()

        self.lam = lam
    def forward(self, img1, img2):
        (_, channel, _, _) = img1.size()

        if channel == self.channel and self.window.dtype == img1.dtype:
            window = self.window
        else:
            window = create_window(self.window_size, channel).to(img1.device).type(img1.dtype)
            self.window = window
            self.channel = channel

        s_score, ssim_map = ssim(img1, img2, window=window, window_size=self.window_size, size_average=self.size_average)
        loss = (1.0 - s_score) * self.lam
        return loss

@LOSS.register_module
class FFTLoss(nn.Module):
    def __init__(self):
        super(FFTLoss, self).__init__()

    def forward(self, input):
        loss = torch.fft.fft2(input).abs().mean()
        return loss


@LOSS.register_module
class SumLoss(nn.Module):
    def __init__(self, lam=1):
        super(SumLoss, self).__init__()
        self.lam = lam

    def forward(self, input1, input2):
        input1 = input1 if isinstance(input1, list) else [input1]
        input2 = input2 if isinstance(input2, list) else [input2]
        loss = 0
        for in1, in2 in zip(input1, input2):
            loss += (in1 + in2) * self.lam
        return loss


@LOSS.register_module
class CSUMLoss(nn.Module):
    def __init__(self, lam=1):
        super(CSUMLoss, self).__init__()
        self.lam = lam

    def forward(self, input):
        loss = 0
        for instance in input:
            _, _, h, w = instance.shape
            loss += torch.sum(instance) / (h * w) * self.lam
        return loss

@LOSS.register_module
class FFocalLoss(nn.Module):
    def __init__(self, lam=1, alpha=-1, gamma=4, reduction="mean"):
        super(FFocalLoss, self).__init__()
        self.lam = lam
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction

    def forward(self, inputs, targets):
        inputs = inputs.float()
        targets = targets.float()
        ce_loss = F.binary_cross_entropy(inputs, targets, reduction="none")
        p_t = inputs * targets + (1 - inputs) * (1 - targets)
        loss = ce_loss * ((1 - p_t) **  self.gamma)

        if  self.alpha >= 0:
            alpha_t =  self.alpha * targets + (1 -  self.alpha) * (1 - targets)
            loss = alpha_t * loss

        if  self.reduction == "mean":
            loss = loss.mean() * self.lam
        elif  self.reduction == "sum":
            loss = loss.sum() * self.lam

        return loss

@LOSS.register_module
class SegmentCELoss(nn.Module):
    def __init__(self, weight):
        super().__init__()
        self.criterion = nn.CrossEntropyLoss()
        self.weight = weight

    def forward(self, mask, pred):
        bsz,_,h,w=pred.size()
        pred = pred.view(bsz, 2, -1)
        mask = mask.view(bsz, -1).long()
        return self.criterion(pred,mask)


import logging
import os
import cv2
import numpy as np
from typing import List, Dict, Union
import copy
import torch
from tqdm import tqdm
import torch.multiprocessing as mp
from easydict import EasyDict

from hiad.utils.base import create_logger
from hiad.utils.base import split_list, save_tasks_config, load_tasks
from hiad.utils.split_and_gather import NPImageGather, HRSample
from hiad.utils.device import train_in_device, test_in_device, compute_metrics_in_device
from hiad.utils.base import create_share_array, Report
from hiad.task.utils import update_paths_and_indexes

class HRTrainer:
    def __init__(self,
                 detector_class,
                 config,
                 batch_size: int,
                 checkpoint_root: str,
                 log_root: str,
                 tasks: List[Dict] = None,
                 vis_root: str = None,
                 seed: int = 0,
                 fusion_weights: List = None,
                 early_stop_epochs = -1):
        r"""
           High-resolution trainer

           Args:
               detector_class (class): The provided detector class template, used to create detector instances across multiple processes.
                    Supported detector types are listed in `detectors.__init__`.
               config (dict): Detector configuration parameters.
               batch_size (size): Batch Size
               checkpoint_root (str): Directory for saving model checkpoints
               log_root (str): Directory for saving Logs
               vis_root (str): Directory for saving visualization results. If None, visualization of results will be ignored.
               tasks (List[Dict]): The provided task list are created by the TaskGenerator. Supported TaskGenerators include:
                                   O2OTaskGenerator, A2OTaskGenerator, NeighborhoodTaskGenerator, SpatialClusteringTaskGenerator,
                                   ThresholdClusteringTaskGenerator, and RetrieveTaskGenerator.
                                   If None, it will be loaded from the checkpoint.
               seed (int): random seed.
               fusion_weights (list): Weights used for feature fusion. If None, a set of equal weights will be used.
               early_stop_epochs (int): Controls early stopping for detector: if a detector shows no performance improvement on the validation set for N epochs, training will be stopped.
                                        If -1, Early stopping will be disabled.
        """

        self.detector_class = detector_class
        self.batch_size = batch_size
        self.checkpoint_root = checkpoint_root
        self.log_root = log_root
        self.vis_root = vis_root
        self.seed = seed
        self.tasks = tasks
        self.early_stop_epochs = early_stop_epochs
        self.fusion_weights = fusion_weights

        if type(config) == dict:
            config = EasyDict(config)
        self.config = config
        self.use_thumbnail = 'thumbnail' in self.config
        mp.set_start_method('spawn', force=True)

    def train(self,
              train_samples: List[HRSample],
              gpu_ids: List[int],
              main_logger = None,
              val_config: Dict = None):
        r"""
         Args:
             train_samples (List[HRSample]): List of provided high-resolution training samples.
             gpu_ids (List[int]): List of GPUs id.
             main_logger (logging.Logger): Logger object to use. If None, a new one will be created.
             val_config (dict): This configuration is used to split the validation set, used for checkpoint selection and normalization.
                For example:
                    val_config = {
                            'anomaly_syn_fn': RandomeBoxSynthesizer(p=0.5), # Defines the anomaly synthesis method used to create anomaly samples.
                            'val_sample_ratio': 0.2, # Number of samples from `train_samples` to be used as the validation set. (0.2 means 20%)
                            'max_patch_number': 200, # Maximum number of validation patches per task. The default value is -1, which means no limit.
                            'evaluators': evaluators # Validation set evaluation methods.
                        }
                If None, validation will be disabled.
        """
        assert self.tasks is not None, 'tasks is None'

        if main_logger is None:
            main_logger = create_logger("main", os.path.join(self.log_root, 'main.log'), print_console=True)

        main_logger.info(f'Start training, devices: {gpu_ids}')

        if self.use_thumbnail:
            thumbnail_task = {'name': 'thumbnail', 'thumbnail_size': self.config.thumbnail.thumbnail_size}
            self.tasks.append(thumbnail_task)
            main_logger.info('Add thumbnail task: {}'.format(thumbnail_task))

        tasks_save_path = os.path.join(self.checkpoint_root, 'tasks.json')
        save_tasks_config(self.tasks, tasks_save_path)
        main_logger.info(f'Tasks config is saved as: {tasks_save_path}')

        for index, task in enumerate(self.tasks):
            if task['name'].startswith('retrieve'):
                main_logger.info(f"[{index+1}/{len(self.tasks)}] Task: {task['name']}, patch_number: {len(task['paths_and_indexes'])}")
            elif task['name'] == 'thumbnail':
                main_logger.info(f"[{index+1}/{len(self.tasks)}] Task: {task['name']}, thumbnail_size: {task['thumbnail_size']}")
            else:
                main_logger.info(f"[{index+1}/{len(self.tasks)}] Task: {task['name']}, Indexes: { [ str(index) for index in task['indexes']]}")

        main_logger.info(f'The training progress can be monitored in: {self.log_root}.')

        num_device = len(gpu_ids)
        tasks_in_device = split_list(self.tasks, num_device)

        results = []
        process_pool = mp.Pool(processes = num_device)
        for gpu_id, task in zip(gpu_ids, tasks_in_device):
            results.append(process_pool.apply_async(train_in_device, args=(gpu_id, self.detector_class, self.config,
                            copy.deepcopy(train_samples), task, self.batch_size,
                            self.checkpoint_root, self.log_root, self.seed, copy.deepcopy(val_config), self.fusion_weights, self.early_stop_epochs)))

        process_pool.close()
        process_pool.join()
        for result in results:
            result = result.get()
            if result is not None:
                main_logger.info(result)

        main_logger.info(f'End training')
        main_logger.info(f'Checkpoints are saved as: {self.checkpoint_root}')


    def inference(self,
                  test_samples: List[HRSample],
                  gpu_ids: List[int],
                  return_results_only = False,
                  evaluators: List = None,
                  main_logger = None,
                  vis_size: Union[int, List] = 1024):

        r"""
              Args:
                  test_samples (List[HRSample]): List of provided high-resolution testing samples.
                  gpu_ids (List[int]): List of GPUs id.
                  main_logger (logging.Logger): Logger object to use. If None, a new one will be created.
                  return_results_only (bool): Return results only. Default: False
                  evaluators: evaluation methods. Setting this to None is only allowed when `return_results_only` is set to True.
                  vis_size: Resolution of the saved visualization images; Default: 1024×1024.
        """

        if os.path.exists(os.path.join(self.checkpoint_root, 'tasks.json')):
            self.tasks = load_tasks(os.path.join(self.checkpoint_root, 'tasks.json'))
            if not self.use_thumbnail:
                self.tasks = [task for task in self.tasks if task['name'] != 'thumbnail']

        if main_logger is None:
            main_logger = create_logger("main", os.path.join(self.log_root, 'main.log'), print_console=True)

        main_logger.info(f'Start inference, devices: {gpu_ids}')
        main_logger.info(f'The inference progress can be monitored in : {self.log_root}.')

        num_device = len(gpu_ids)
        retrieve_tasks = [task for task in self.tasks if task['name'].startswith('retrieve')]

        if len(retrieve_tasks) != 0:
            main_logger.info(f'Start Retrieve')
            retrieve_tasks = update_paths_and_indexes(
                                     copy.deepcopy(retrieve_tasks),
                                     test_samples,
                                     self.detector_class,
                                     self.config.patch,
                                     self.batch_size,
                                     gpu_ids)

            tasks = [task for task in retrieve_tasks if len(task['paths_and_indexes']) != 0] + [task for task in self.tasks if not task['name'].startswith('retrieve')]
            main_logger.info(f'Retrieve End')
        else:
            tasks = self.tasks

        process_pool = mp.Pool(processes=num_device)
        tasks_in_device = split_list(tasks, num_device)

        device_results = []
        for gpu_id, task in zip(gpu_ids, tasks_in_device):
            device_result = process_pool.apply_async(test_in_device, (gpu_id, self.detector_class, self.config,
                                                            copy.deepcopy(test_samples), task, self.batch_size,
                                                            self.checkpoint_root, self.log_root, self.seed))
            device_results.append(device_result)

        process_pool.close()
        process_pool.join()

        prediction_masks = {sample.image.image_path: {"patch_list": [], "index_list": []} for sample in test_samples}

        for device_result in device_results:
            device_result = device_result.get()
            for key in device_result:
                prediction_masks[key]['patch_list'].extend(device_result[key]['patch_list'])
                prediction_masks[key]['index_list'].extend(device_result[key]['index_list'])
                if 'thumbnail' in device_result[key]:
                    prediction_masks[key]['thumbnail'] = device_result[key]['thumbnail']

        patch_prediction_masks = {}
        thumbnail_prediction_masks = {}

        for key in tqdm(prediction_masks, desc='Gather Segmentations'):
            segmentation = NPImageGather(patch_list = prediction_masks[key]['patch_list'],
                                         index_list = prediction_masks[key]['index_list'])
            patch_prediction_masks[key] = segmentation
            if 'thumbnail' in prediction_masks[key]:
                thumbnail_prediction_masks[key] = prediction_masks[key]['thumbnail']

        patch_prediction_masks = np.stack([patch_prediction_masks[sample.image.image_path] for sample in test_samples])

        if thumbnail_prediction_masks != {}:
            thumbnail_prediction_masks = np.stack(
                [thumbnail_prediction_masks[sample.image.image_path] for sample in test_samples])
            assert thumbnail_prediction_masks.shape[0] == patch_prediction_masks.shape[0]
        else:
            thumbnail_prediction_masks = None

        prediction_masks = self.detector_class.post_processing(patch_prediction_masks, thumbnail_prediction_masks,
                                                               is_normalize= not return_results_only,
                                                               gaussian_filter = True)
        prediction_scores = self.detector_class.get_image_score(prediction_masks, batch_size = self.batch_size)

        if return_results_only:
            return prediction_scores, [prediction_masks[i] for i in range(prediction_masks.shape[0])]

        assert evaluators is not None, "No evaluation function provided."

        gt_masks = []
        gt_labels = []

        for sample in tqdm(test_samples, desc='Loading GT Masks'):
            if sample.mask is not None:
                sample.open()
                mask = copy.deepcopy(sample.mask.image)
                mask[mask != 0] = 1
                sample.close()
            else:
                mask = np.zeros(prediction_masks.shape[-2:]).astype(float)

            if sample.label is not None:
                gt_labels.append(sample.label)
            else:
                gt_labels.append(np.max(mask).item())

            gt_masks.append(mask)

        gt_masks = np.stack(gt_masks)

        clsname_list = [sample.clsname if sample.clsname is not None else 'unknown' for sample in test_samples]
        all_clsnames = list(set(clsname_list))

        torch.cuda.empty_cache()
        main_logger.info(f'Computing Metrics')

        device_results = []

        process_pool = mp.Pool(processes=num_device)
        clsnames_in_device = split_list(all_clsnames, num_device)

        prediction_masks = create_share_array(prediction_masks)
        gt_masks = create_share_array(gt_masks)

        for gpu_id, clsnames in zip(gpu_ids, clsnames_in_device):
            device_result = process_pool.apply_async(compute_metrics_in_device, (gpu_id, clsnames, evaluators,
                                                                                 prediction_masks, gt_masks,
                                                                                 gt_labels, prediction_scores,
                                                                                 clsname_list))
            device_results.append(device_result)

        process_pool.close()
        process_pool.join()

        scores = []
        for device_result in device_results:
            cls_score = device_result.get()
            scores.extend(cls_score)

        scores.sort(key = lambda a: a['clsname'])

        all_metrics = [item for item in scores[0] if item.find('threshold') == -1 and item.find('clsname') == -1]

        record = Report(["clsname"] + all_metrics)

        for score in scores:
            clsvalues = [
                score[metric] for metric in all_metrics
            ]
            record.add_one_record([score['clsname']] + clsvalues)

        mean_metrics = []
        for metric in all_metrics:
            mean_metrics.append(np.mean([score[metric] for score in scores]))

        record.add_one_record(['mean'] + mean_metrics)

        main_logger.info(f"\n{record}")

        if self.vis_root is not None:

            vis_size = (vis_size, vis_size) if isinstance(vis_size, int) else vis_size

            from skimage.segmentation import mark_boundaries
            from PIL import Image
            from pytorch_grad_cam.utils.image import show_cam_on_image

            main_logger.info(f'Starting save Vis ...')

            if 'seg_threshold' in scores[0]:
                seg_thresholds = {score['clsname']: score['seg_threshold'] for score in scores}
            else:
                seg_thresholds = None

            prediction_masks = prediction_masks.numpy() if isinstance(prediction_masks, torch.Tensor) else prediction_masks
            gt_masks = gt_masks.numpy() if isinstance(gt_masks, torch.Tensor) else gt_masks

            for index, (sample, pred_mask, gt_mask, clsname) in enumerate(tqdm(zip(test_samples, prediction_masks, gt_masks, clsname_list), total=len(test_samples))):
                sample.open()

                image = sample.down_sampling_to_LR(vis_size).image
                pred_mask = cv2.resize(pred_mask, vis_size, interpolation=cv2.INTER_NEAREST)
                gt_mask = cv2.resize(gt_mask,  vis_size, interpolation=cv2.INTER_NEAREST)

                heat = show_cam_on_image(image / 255, pred_mask, use_rgb=True)

                if seg_thresholds is not None:
                    pred_mask[pred_mask >= seg_thresholds[clsname]] = 1
                    pred_mask[pred_mask < seg_thresholds[clsname]] = 0
                    image_with_pred = mark_boundaries(image / 255, pred_mask, color=(1, 0, 0), mode='inner')
                    image_with_mask = mark_boundaries(image / 255, gt_mask, color=(1, 0, 0), mode='inner')
                    save_image = np.concatenate([image_with_pred * 255, heat, image_with_mask * 255], axis=1)
                else:
                    image_with_mask = mark_boundaries(image / 255, gt_mask, color=(1, 0, 0), mode='inner')
                    save_image = np.concatenate([image, heat, image_with_mask * 255], axis=1)

                _, image_name = os.path.split(sample.image.image_path)
                Image.fromarray(save_image.astype(np.uint8)).save(os.path.join(self.vis_root, "{}{}_{}".format(clsname, index, image_name)))
                sample.close()

        main_logger.info(f'End Inference')



import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as T
import torchvision.transforms.functional as F_tv
from timm.data.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from . import LOSS

__all__ = ['CE', 'LabelSmoothingCE', 'SoftTargetCE']


@LOSS.register_module
class CE(nn.CrossEntropyLoss):
	def __init__(self, lam=1):
		super(CE, self).__init__()
		self.lam = lam

	def forward(self, input, target):
		return super(CE, self).forward(input, target) * self.lam


@LOSS.register_module
class LabelSmoothingCE(nn.Module):
	"""
	NLL loss with label smoothing.
	"""
	def __init__(self, smoothing=0.1, lam=1):
		"""
		Constructor for the LabelSmoothing module.
		:param smoothing: label smoothing factor
		"""
		super(LabelSmoothingCE, self).__init__()
		assert smoothing < 1.0
		self.smoothing = smoothing
		self.lam = lam
		self.confidence = 1. - smoothing

	def forward(self, x, target):
		logprobs = F.log_softmax(x, dim=-1)
		nll_loss = -logprobs.gather(dim=-1, index=target.unsqueeze(1))
		nll_loss = nll_loss.squeeze(1)
		smooth_loss = -logprobs.mean(dim=-1)
		loss = self.confidence * nll_loss + self.smoothing * smooth_loss
		return loss.mean() * self.lam


@LOSS.register_module
class SoftTargetCE(nn.Module):
	def __init__(self, lam=1, fp32=False):
		super(SoftTargetCE, self).__init__()
		self.lam = lam
		self.fp32 = fp32

	def forward(self, x, target):
		if self.fp32:
			loss = torch.sum(-target * F.log_softmax(x.float(), dim=-1), dim=-1)
		else:
			loss = torch.sum(-target * F.log_softmax(x, dim=-1), dim=-1)
		return loss.mean() * self.lam




import numpy as np
import torch
import torch.nn as nn
try:
	from torch.hub import load_state_dict_from_url
except ImportError:
	from torch.utils.model_zoo import load_url as load_state_dict_from_url
from timm.models.vision_transformer import VisionTransformer, _cfg, checkpoint_filter_fn
from timm.models.helpers import build_model_with_cfg
from timm.models._manipulate import  checkpoint_seq
from functools import partial
from timm.models.layers import trunc_normal_
from ..util.net import Registry


MODEL = Registry('Model')


def get_model(cfg_model):
	model_name = cfg_model.name
	kwargs = {k: v for k, v in cfg_model.kwargs.items()}
	model_fn = MODEL.get_module(model_name)
	pretrained = kwargs.pop('pretrained')
	model = model_fn(pretrained=pretrained, **kwargs)
	return model


class Fusion(nn.Module):
	def __init__(self, dim, mul):
		super(Fusion, self).__init__()
		self.fc = nn.Linear(dim * mul, dim)

	def forward(self, features):
		# B, L, C
		feature_align = torch.cat(features, dim=2)
		feature_align = self.fc(feature_align)
		return feature_align


@MODEL.register_module
def fusion(pretrained=False, **kwargs):
	model = Fusion(**kwargs)
	return model

# ========== ViT Encoder ==========
# deit encoder
class DistilledVisionTransformer(VisionTransformer):
	def __init__(self, teachers, neck, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.teachers = teachers
		self.neck = neck
		self.dist_token = nn.Parameter(torch.zeros(1, 1, self.embed_dim))
		num_patches = self.patch_embed.num_patches
		self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 2, self.embed_dim))
		self.head_dist = nn.Linear(self.embed_dim, self.num_classes) if self.num_classes > 0 else nn.Identity()

		trunc_normal_(self.dist_token, std=.02)
		trunc_normal_(self.pos_embed, std=.02)
		self.head_dist.apply(self._init_weights)

	def forward(self, x):
		B = x.shape[0]
		x = self.patch_embed(x)

		cls_tokens = self.cls_token.expand(B, -1, -1)  # stole cls_tokens impl from Phil Wang, thanks
		dist_token = self.dist_token.expand(B, -1, -1)
		x = torch.cat((cls_tokens, dist_token, x), dim=1)

		x = x + self.pos_embed
		x = self.pos_drop(x)

		out, neck = [], []
		for idx, blk in enumerate(self.blocks):
			x = blk(x)
			fea = x[:, 2:]
			B, L, C = fea.shape
			H = int(np.sqrt(L))
			fea = fea.view(B, H, H, C).permute(0, 3, 1, 2).contiguous()  # B, C, H, W

			if (idx + 1) in self.neck:
				neck.append(fea)  # B, L, C
			if (idx + 1) in self.teachers:
				out.append(fea)
		return out, neck

# normal encoder
class ViT_Encoder(VisionTransformer):
	def __init__(self, teachers, neck, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.teachers = teachers
		self.neck = neck

	def forward(self, x):
		x = self.patch_embed(x)
		x = self._pos_embed(x)

		out_neck, out_t = [], []
		if self.grad_checkpointing and not torch.jit.is_scripting():
			x = checkpoint_seq(self.blocks, x)
		else:
			for i in range(len(self.blocks)):
				x = self.blocks[i](x)
				fea = x[:, 1:, :]
				B, L, C = fea.shape
				H = int(np.sqrt(L))
				fea = fea.view(B, H, H, C).permute(0, 3, 1, 2).contiguous()
				if i+1 in self.neck:
					out_neck.append(fea)
				if i+1 in self.teachers:
					out_t.append(fea)
		return out_t, out_neck

# ========== ViT Decoder ==========
class ViT_Decoder(VisionTransformer):
	def __init__(self, students, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.students = students
		num_patches = self.patch_embed.num_patches
		self.pos_embed = nn.Parameter(torch.zeros(1, num_patches, self.embed_dim))
		trunc_normal_(self.pos_embed, std=.02)

	def forward(self, x):
		# taken from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py
		# B, C, H, W = x.shape
		# x = x.view(B, C, -1).permute(0, 2, 1).contiguous()  # B, L, C
		x = x + self.pos_embed
		x = self.pos_drop(x)

		out = []
		for idx, blk in enumerate(self.blocks):
			x = blk(x)
			if (idx + 1) in self.students:
				fea = x
				B, L, C = fea.shape
				H = int(np.sqrt(L))
				fea = fea.view(B, H, H, C).permute(0, 3, 1, 2).contiguous()
				out.append(fea)

		return [out[int(len(out)-1-i)] for i in range(len(out))]

# ========== Encoders with Different Pre-trained Weights ==========

def _create_vision_transformer(variant, pretrained=False, **kwargs):
	if kwargs.get('features_only', None):
		raise RuntimeError('features_only not implemented for Vision Transformer models.')

	if 'flexi' in variant:
		_filter_fn = partial(checkpoint_filter_fn, interpolation='bilinear', antialias=False)
	else:
		_filter_fn = checkpoint_filter_fn

	model = build_model_with_cfg(ViT_Encoder, variant, pretrained, pretrained_filter_fn=_filter_fn, **kwargs)
	return model


@MODEL.register_module
def vit_small_patch16_224_1k(pretrained=True, **kwargs):
	""" ViT-Small (ViT-S/16)"""
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch16_224.augreg_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_small_patch16_384(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch16_384.augreg_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_small_patch32_224(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=32, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch32_224.augreg_in21k_ft_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_small_patch16_224(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch16_224.augreg_in21k_ft_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_small_patch8_224(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=8, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch8_224.augreg_in21k_ft_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_base_patch32_224(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=32, embed_dim=768, depth=12, num_heads=12)
	model = _create_vision_transformer('vit_base_patch32_224.augreg_in21k_ft_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_base_patch16_384(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=12, num_heads=12)
	model = _create_vision_transformer('vit_base_patch16_384.augreg_in21k_ft_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_base_patch16_224(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=12, num_heads=12)
	model = _create_vision_transformer('vit_base_patch16_224.augreg_in21k_ft_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_base_patch8_224(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=8, embed_dim=768, depth=12, num_heads=12)
	model = _create_vision_transformer('vit_base_patch8_224.augreg_in21k_ft_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_large_patch16_224(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=1024, depth=24, num_heads=16)
	model = _create_vision_transformer('vit_large_patch16_224.augreg_in21k_ft_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_huge_patch14_224(pretrained=True, **kwargs):
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=14, embed_dim=1280, depth=32, num_heads=16)
	model = _create_vision_transformer('vit_huge_patch14_224.orig_in21k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model

### DINO
@MODEL.register_module
def vit_small_patch16_224_dino(pretrained=True, **kwargs):
	""" ViT-Small-DINO, patch16"""
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch16_224.dino', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model

@MODEL.register_module
def vit_small_patch8_224_dino(pretrained=True, **kwargs):
	""" ViT-Small-DINO, patch8"""
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=8, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch8_224.dino', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


@MODEL.register_module
def vit_base_patch16_224_dino(pretrained=True, **kwargs):
	""" ViT-Base-DINO, patch16"""
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=12, num_heads=12)
	model = _create_vision_transformer('vit_base_patch16_224.dino', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model


### DINOv2
@MODEL.register_module
def vit_small_patch16_224_dino2(pretrained=True, **kwargs):
	""" ViT-Small-DINO, patch16"""
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch14_dinov2.lvd142m', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model

@MODEL.register_module
def vit_small_patch16_224_dino2reg(pretrained=True, **kwargs):
	""" ViT-Small-DINO, patch16"""
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=12, num_heads=6)
	model = _create_vision_transformer('vit_small_patch14_reg4_dinov2.lvd142m', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model

### DeiT
@MODEL.register_module
def deit_tiny_distilled_patch16_224(pretrained=True, **kwargs):
	model = DistilledVisionTransformer(
        img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=192, depth=12, num_heads=3, mlp_ratio=4, qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
	model.default_cfg = _cfg()
	if pretrained:
		checkpoint = torch.hub.load_state_dict_from_url(
            url="https://dl.fbaipublicfiles.com/deit/deit_tiny_distilled_patch16_224-b40b3cf7.pth",
            map_location="cpu", check_hash=True
        )
		model.load_state_dict(checkpoint["model"])
	return model

@MODEL.register_module
def deit_small_distilled_patch16_224(pretrained=True, **kwargs):
	model = DistilledVisionTransformer(
		img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=12, num_heads=6, mlp_ratio=4, qkv_bias=True,
		norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
	model.default_cfg = _cfg()
	if pretrained:
		checkpoint = torch.hub.load_state_dict_from_url(
			url="https://dl.fbaipublicfiles.com/deit/deit_small_distilled_patch16_224-649709d9.pth",
			map_location="cpu", check_hash=True
		)
		model.load_state_dict(checkpoint["model"])
	return model


### CLIP
@MODEL.register_module
def vit_base_patch16_clip_224(pretrained=True, **kwargs):
	""" ViT-B/16 CLIP image tower"""
	model_kwargs = dict(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=12, num_heads=12,
						pre_norm=True, norm_layer=nn.LayerNorm)
	model = _create_vision_transformer(
		'vit_base_patch16_clip_224.openai_ft_in12k_in1k', pretrained=pretrained, **dict(model_kwargs, **kwargs))
	return model

# ========== Decoders with Different Pre-trained Weights ==========
### ViT
@MODEL.register_module
def de_vit_small_patch16_224_1k(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=kwargs.pop('depth'), num_heads=6, **kwargs)
	return model


@MODEL.register_module
def de_vit_small_patch16_384(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=kwargs.pop('depth'), num_heads=6, **kwargs)
	return model

@MODEL.register_module
def de_vit_small_patch16_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=kwargs.pop('depth'), num_heads=6, **kwargs)
	return model

@MODEL.register_module
def de_vit_small_patch8_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=8, embed_dim=384, depth=kwargs.pop('depth'), num_heads=6, **kwargs)
	return model

@MODEL.register_module
def de_vit_small_patch32_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=32, embed_dim=384, depth=kwargs.pop('depth'), num_heads=6, **kwargs)
	return model


@MODEL.register_module
def de_vit_base_patch16_384(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=kwargs.pop('depth'), num_heads=12, **kwargs)
	return model


@MODEL.register_module
def de_vit_base_patch16_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=kwargs.pop('depth'), num_heads=12, **kwargs)
	return model

@MODEL.register_module
def de_vit_base_patch8_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=8, embed_dim=768, depth=kwargs.pop('depth'), num_heads=12, **kwargs)
	return model


@MODEL.register_module
def de_vit_base_patch32_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=32, embed_dim=768, depth=kwargs.pop('depth'), num_heads=12, **kwargs)
	return model


@MODEL.register_module
def de_vit_large_patch16_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=1024, depth=kwargs.pop('depth'), num_heads=16, **kwargs)
	return model

@MODEL.register_module
def de_vit_huge_patch14_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=14, embed_dim=1280, depth=kwargs.pop('depth'), num_heads=16, **kwargs)
	return model

### DINO
@MODEL.register_module
def de_vit_small_patch16_224_dino(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=kwargs.pop('depth'), num_heads=6, **kwargs)
	return model


@MODEL.register_module
def de_vit_base_patch16_224_dino(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=kwargs.pop('depth'), num_heads=12, **kwargs)
	return model

@MODEL.register_module
def de_vit_small_patch8_224_dino(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=8, embed_dim=384, depth=kwargs.pop('depth'), num_heads=6, **kwargs)
	return model


### DeiT
@MODEL.register_module
def de_deit_tiny_distilled_patch16_224(pretrained=False, **kwargs):
	model = ViT_Decoder(
		img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=192, depth=kwargs.pop('depth'), num_heads=3,
		mlp_ratio=4, qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
	return model

@MODEL.register_module
def de_deit_small_distilled_patch16_224(pretrained=False, **kwargs):
	model = ViT_Decoder(
		img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=384, depth=kwargs.pop('depth'), num_heads=6,
		mlp_ratio=4, qkv_bias=True, norm_layer=partial(nn.LayerNorm, eps=1e-6), **kwargs)
	return model

### MAE
@MODEL.register_module
def de_mae_vit_base_patch16_dec512d8b(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=kwargs.pop('depth'), num_heads=12, **kwargs)
	return model

### CLIP
@MODEL.register_module
def de_vit_base_patch16_clip_224(pretrained=False, **kwargs):
	model = ViT_Decoder(img_size=kwargs.pop('img_size'), patch_size=16, embed_dim=768, depth=kwargs.pop('depth'), num_heads=12, **kwargs)
	return model


# ========== ViTAD ==========
class ViTAD(nn.Module):
	def __init__(self, model_t, model_f, model_s):
		super(ViTAD, self).__init__()
		self.net_t = get_model(model_t)
		self.net_fusion = get_model(model_f)
		self.net_s = get_model(model_s)
		self.frozen_layers = ['net_t']

	def freeze_layer(self, module):
		module.eval()
		for param in module.parameters():
			param.requires_grad = False

	def train(self, mode=True):
		self.training = mode
		for mname, module in self.named_children():
			if mname in self.frozen_layers:
				self.freeze_layer(module)
			else:
				module.train(mode)
		return self

	def forward(self, imgs):
		feats_t, feats_n = self.encoder(imgs)
		feats_t, feats_s = self.decoder(feats_t, feats_n)
		return feats_t, feats_s

	def encoder(self, imgs):
		feats_t, feats_n = self.net_t(imgs)  # list
		feats_t = [f.detach() for f in feats_t]
		feats_n = [f.detach() for f in feats_n]
		return feats_t, feats_n

	def decoder(self, feats_t, feats_n):
		feats_s = self.net_s(self.net_fusion(feats_n))
		return feats_t, feats_s



@MODEL.register_module
def vitad(pretrained=False, **kwargs):
	model = ViTAD(**kwargs)
	return model


from collections.abc import Callable
from typing import Any
import torch
from torchmetrics import Metric
from torchmetrics.functional.classification import binary_roc
from torchmetrics.utilities.compute import auc
from torchmetrics.utilities.data import dim_zero_cat
from torch import linspace
import cv2
import numpy as np
from kornia.contrib import connected_components


def thresholds_between_min_and_max(
    preds: torch.Tensor,
    num_thresholds: int = 100,
    device: torch.device  = None,
) -> torch.Tensor:
    """Threshold values between min and max of the predictions.

    Args:
        preds (torch.Tensor): Predictions.
        num_thresholds (int, optional): Number of thresholds to generate. Defaults to 100.
        device (torch_device | None, optional): Device to use for computation. Defaults to None.

    Returns:
        Tensor:
            Array of size ``num_thresholds`` that contains evenly spaced values
            between ``preds.min()`` and ``preds.max()`` on ``device``.
    """
    return linspace(start=preds.min(), end=preds.max(), steps=num_thresholds, device=device)



def thresholds_between_0_and_1(num_thresholds: int = 100, device: torch.device  = None) -> torch.Tensor:
    """Threshold values between 0 and 1.

    Args:
        num_thresholds (int, optional): Number of thresholds to generate. Defaults to 100.
        device (torch_device | None, optional): Device to use for computation. Defaults to None.

    Returns:
        Tensor: Threshold values between 0 and 1.
    """
    return linspace(start=0, end=1, steps=num_thresholds, device=device)



def connected_components_gpu(image: torch.Tensor, num_iterations: int = 1000) -> torch.Tensor:
    """Perform connected component labeling on GPU and remap the labels from 0 to N.

    Args:
        image (torch.Tensor): Binary input image from which we want to extract connected components (Bx1xHxW)
        num_iterations (int): Number of iterations used in the connected component computation.

    Returns:
        Tensor: Components labeled from 0 to N.
    """
    components = connected_components(image, num_iterations=num_iterations)

    # remap component values from 0 to N
    labels = components.unique()
    for new_label, old_label in enumerate(labels):
        components[components == old_label] = new_label

    return components.int()


def connected_components_cpu(image: torch.Tensor) -> torch.Tensor:
    """Perform connected component labeling on CPU.

    Args:
        image (torch.Tensor): Binary input data from which we want to extract connected components (Bx1xHxW)

    Returns:
        Tensor: Components labeled from 0 to N.
    """
    components = torch.zeros_like(image)
    label_idx = 1
    for i, msk in enumerate(image):
        mask = msk.squeeze().cpu().numpy().astype(np.uint8)
        _, comps = cv2.connectedComponents(mask)
        # remap component values to make sure every component has a unique value when outputs are concatenated
        for label in np.unique(comps)[1:]:
            components[i, 0, ...][np.where(comps == label)] = label_idx
            label_idx += 1
    return components.int()



class AUPRO(Metric):
    """Area under per region overlap (AUPRO) Metric.

    Args:
        dist_sync_on_step (bool): Synchronize metric state across processes at each ``forward()``
            before returning the value at the step. Default: ``False``
        process_group (Optional[Any]): Specify the process group on which synchronization is called.
            Default: ``None`` (which selects the entire world)
        dist_sync_fn (Optional[Callable]): Callback that performs the allgather operation on the metric state.
            When ``None``, DDP will be used to perform the allgather.
            Default: ``None``
        fpr_limit (float): Limit for the false positive rate. Defaults to ``0.3``.
        num_thresholds (int): Number of thresholds to use for computing the roc curve. Defaults to ``None``.
            If ``None``, the roc curve is computed with the thresholds returned by
            ``torchmetrics.functional.classification.thresholds``.

    Examples:
        >>> import torch
        ...
        >>> labels = torch.randint(low=0, high=2, size=(1, 10, 5), dtype=torch.float32)
        >>> preds = torch.rand_like(labels)
        ...
        >>> aupro = AUPRO(fpr_limit=0.3)
        >>> aupro(preds, labels)
        tensor(0.4321)

        Increasing the fpr_limit will increase the AUPRO value:

        >>> aupro = AUPRO(fpr_limit=0.7)
        >>> aupro(preds, labels)
        tensor(0.5271)
    """

    is_differentiable: bool = False
    higher_is_better: bool  = None
    full_state_update: bool = False
    preds: torch.Tensor = None
    target: torch.Tensor = None
    # When not None, the computation is performed in constant-memory by computing the roc curve
    # for fixed thresholds buckets/thresholds.
    # Warning: The thresholds are evenly distributed between the min and max predictions
    # if all predictions are inside [0, 1]. Otherwise, the thresholds are evenly distributed between 0 and 1.
    # This warning can be removed when https://github.com/Lightning-AI/torchmetrics/issues/1526 is fixed
    # and the roc curve is computed with deactivated formatting
    num_thresholds: int = None

    def __init__(
            self,
            dist_sync_on_step: bool = False,
            process_group: Any  = None,  # noqa: ANN401
            dist_sync_fn: Callable  = None,
            fpr_limit: float = 0.3,
            num_thresholds: int  = None,
    ) -> None:
        super().__init__(
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

        self.add_state("preds", default=[], dist_reduce_fx="cat")
        self.add_state("target", default=[], dist_reduce_fx="cat")
        self.register_buffer("fpr_limit", torch.tensor(fpr_limit))
        self.num_thresholds = num_thresholds

    def update(self, preds: torch.Tensor, target: torch.Tensor) -> None:
        """Update state with new values.

        Args:
            preds (torch.Tensor): predictions of the model
            target (torch.Tensor): ground truth targets
        """
        self.target.append(target)
        self.preds.append(preds)

    def perform_cca(self) -> torch.Tensor:
        """Perform the Connected Component Analysis on the self.target tensor.

        Raises:
            ValueError: ValueError is raised if self.target doesn't conform with requirements imposed by kornia for
                        connected component analysis.

        Returns:
            Tensor: Components labeled from 0 to N.
        """
        target = dim_zero_cat(self.target)

        # check and prepare target for labeling via kornia
        if target.min() < 0 or target.max() > 1:
            msg = (
                "kornia.contrib.connected_components expects input to lie in the interval [0, 1], "
                f"but found interval was [{target.min()}, {target.max()}]."
            )
            raise ValueError(
                msg,
            )
        target = target.unsqueeze(1)  # kornia expects N1HW format
        target = target.type(torch.float)  # kornia expects FloatTensor
        return  connected_components_cpu(target)

    def compute_pro(
            self,
            cca: torch.Tensor,
            target: torch.Tensor,
            preds: torch.Tensor,
    ) :
        """Compute the pro/fpr value-pairs until the fpr specified by self.fpr_limit.

        It leverages the fact that the overlap corresponds to the tpr, and thus computes the overall
        PRO curve by aggregating per-region tpr/fpr values produced by ROC-construction.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: tuple containing final fpr and tpr values.
        """
        if self.num_thresholds is not None:
            # binary_roc is applying a sigmoid on the predictions before computing the roc curve
            # when some predictions are out of [0, 1], the binning between min and max predictions
            # cannot be applied in that case. This can be removed when
            #  https://github.com/Lightning-AI/torchmetrics/issues/1526 is fixed and
            #  the roc curve is computed with deactivated formatting.

            if torch.all((preds >= 0) * (preds <= 1)):
                thresholds = thresholds_between_min_and_max(preds, self.num_thresholds, preds.device)
            else:
                thresholds = thresholds_between_0_and_1(self.num_thresholds, preds.device)
        else:
            thresholds = None

        # compute the global fpr-size
        fpr: torch.Tensor = binary_roc(
            preds=preds,
            target=target,
            thresholds=thresholds,
        )[0]  # only need fpr

        self.fpr_limit = self.fpr_limit.to(preds.device)

        output_size = torch.where(fpr <= self.fpr_limit)[0].size(0)

        # compute the PRO curve by aggregating per-region tpr/fpr curves/values.
        tpr = torch.zeros(output_size, device=preds.device, dtype=torch.float)
        fpr = torch.zeros(output_size, device=preds.device, dtype=torch.float)
        new_idx = torch.arange(0, output_size, device=preds.device, dtype=torch.float)

        # Loop over the labels, computing per-region tpr/fpr curves, and aggregating them.
        # Note that, since the groundtruth is different for every all to `roc`, we also get
        # different/unique tpr/fpr curves (i.e. len(_fpr_idx) is different for every call).
        # We therefore need to resample per-region curves to a fixed sampling ratio (defined above).
        labels = cca.unique()[1:]  # 0 is background
        background = cca == 0
        _fpr: torch.Tensor
        _tpr: torch.Tensor
        for label in labels:
            interp: bool = False
            new_idx[-1] = output_size - 1
            mask = cca == label
            # Need to calculate label-wise roc on union of background & mask, as otherwise we wrongly consider other
            # label in labels as FPs. We also don't need to return the thresholds
            _fpr, _tpr = binary_roc(
                preds=preds[background | mask],
                target=mask[background | mask],
                thresholds=thresholds,
            )[:-1]

            # catch edge-case where ROC only has fpr vals > self.fpr_limit
            if _fpr[_fpr <= self.fpr_limit].max() == 0:
                _fpr_limit = _fpr[_fpr > self.fpr_limit].min()
            else:
                _fpr_limit = self.fpr_limit

            _fpr_idx = torch.where(_fpr <= _fpr_limit)[0]
            # if computed roc curve is not specified sufficiently close to self.fpr_limit,
            # we include the closest higher tpr/fpr pair and linearly interpolate the tpr/fpr point at self.fpr_limit
            if not torch.allclose(_fpr[_fpr_idx].max(), self.fpr_limit):
                _tmp_idx = torch.searchsorted(_fpr, self.fpr_limit)
                _fpr_idx = torch.cat([_fpr_idx, _tmp_idx.unsqueeze_(0)])
                _slope = 1 - ((_fpr[_tmp_idx] - self.fpr_limit) / (_fpr[_tmp_idx] - _fpr[_tmp_idx - 1]))
                interp = True

            _fpr = _fpr[_fpr_idx]
            _tpr = _tpr[_fpr_idx]

            _fpr_idx = _fpr_idx.float()
            _fpr_idx /= _fpr_idx.max()
            _fpr_idx *= new_idx.max()

            if interp:
                # last point will be sampled at self.fpr_limit
                new_idx[-1] = _fpr_idx[-2] + ((_fpr_idx[-1] - _fpr_idx[-2]) * _slope)

            _tpr = self.interp1d(_fpr_idx, _tpr, new_idx)
            _fpr = self.interp1d(_fpr_idx, _fpr, new_idx)
            tpr += _tpr
            fpr += _fpr

        # Actually perform the averaging
        tpr /= labels.size(0)
        fpr /= labels.size(0)
        return fpr, tpr

    def _compute(self):
        """Compute the PRO curve.

        Perform the Connected Component Analysis first then compute the PRO curve.

        Returns:
            tuple[torch.Tensor, torch.Tensor]: tuple containing final fpr and tpr values.
        """
        cca = self.perform_cca().flatten()
        target = dim_zero_cat(self.target).flatten()
        preds = dim_zero_cat(self.preds).flatten()

        return self.compute_pro(cca=cca, target=target, preds=preds)

    def compute(self) -> torch.Tensor:
        """Fist compute PRO curve, then compute and scale area under the curve.

        Returns:
            Tensor: Value of the AUPRO metric
        """
        fpr, tpr = self._compute()

        aupro = auc(fpr, tpr, reorder=True)
        return aupro / fpr[-1]  # normalize the area


    @staticmethod
    def interp1d(old_x: torch.Tensor, old_y: torch.Tensor, new_x: torch.Tensor) -> torch.Tensor:
        """Interpolate a 1D signal linearly to new sampling points.

        Args:
            old_x (torch.Tensor): original 1-D x values (same size as y)
            old_y (torch.Tensor): original 1-D y values (same size as x)
            new_x (torch.Tensor): x-values where y should be interpolated at

        Returns:
            Tensor: y-values at corresponding new_x values.
        """
        # Compute slope
        eps = torch.finfo(old_y.dtype).eps
        slope = (old_y[1:] - old_y[:-1]) / (eps + (old_x[1:] - old_x[:-1]))

        # Prepare idx for linear interpolation
        idx = torch.searchsorted(old_x, new_x)

        # searchsorted looks for the index where the values must be inserted
        # to preserve order, but we actually want the preceeding index.
        idx -= 1
        # we clamp the index, because the number of intervals = old_x.size(0) -1,
        # and the left neighbour should hence be at most number of intervals -1, i.e. old_x.size(0) - 2
        idx = torch.clamp(idx, 0, old_x.size(0) - 2)

        # perform actual linear interpolation
        return old_y[idx] + slope[idx] * (new_x - old_x[idx])



import glob
import importlib

from ..util.net import Registry

LOSS = Registry('Loss')

from .base_loss import *
from .cls_loss import *
from .gan_loss import *


def get_loss_terms(loss_terms, device='cpu'):
	terms = {}
	for t in loss_terms:
		t = {k: v for k, v in t.items()}
		t_type = t.pop('type')
		t_name = t.pop('name')
		terms[t_name] = LOSS.get_module(t_type)(**t).to(device).eval()
	return terms



import csv
import logging
import os
import random
import matplotlib.pyplot as plt
import numpy as np
import PIL
import torch
import tqdm


LOGGER = logging.getLogger(__name__)


def set_torch_device():
    return torch.device("cuda")


def fix_seeds(seed, with_torch=True, with_cuda=True):
    """Fixed available seeds for reproducibility.

    Args:
        seed: [int] Seed value.
        with_torch: Flag. If true, torch-related seeds are fixed.
        with_cuda: Flag. If true, torch+cuda-related seeds are fixed
    """
    random.seed(seed)
    np.random.seed(seed)
    if with_torch:
        torch.manual_seed(seed)
    if with_cuda:
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True



import math
import timm
import FrEIA.framework as Ff
import FrEIA.modules as Fm
import torch
import torch.nn as nn
import torch.nn.functional as F


class AverageMeter:
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def subnet_conv_func(kernel_size, hidden_ratio):
    def subnet_conv(in_channels, out_channels):
        hidden_channels = int(in_channels * hidden_ratio)
        return nn.Sequential(
            nn.Conv2d(in_channels, hidden_channels, kernel_size, padding="same"),
            nn.ReLU(),
            nn.Conv2d(hidden_channels, out_channels, kernel_size, padding="same"),
        )

    return subnet_conv


def nf_fast_flow(input_chw, conv3x3_only, hidden_ratio, flow_steps, clamp=2.0):
    nodes = Ff.SequenceINN(*input_chw)
    for i in range(flow_steps):
        if i % 2 == 1 and not conv3x3_only:
            kernel_size = 1
        else:
            kernel_size = 3
        nodes.append(
            Fm.AllInOneBlock,
            subnet_constructor=subnet_conv_func(kernel_size, hidden_ratio),
            affine_clamping=clamp,
            permute_soft=False,
        )
    return nodes


class FastFlow(nn.Module):

    def __init__(
        self,
        backbone_name,
        input_size,
        flow_steps,
        channels,
        scales,
        conv3x3_only=False,
        hidden_ratio=1.0,
    ):

        super(FastFlow, self).__init__()
        self.backbone_name = backbone_name
        self.input_size = input_size
        self.flow_steps = flow_steps
        self.channels = channels
        self.scales = scales
        self.conv3x3_only = conv3x3_only
        self.hidden_ratio = hidden_ratio

        if self.backbone_name.find('resnet') != -1:
            self.norms = nn.ModuleList()
            for in_channels, scale in zip(self.channels, self.scales):
                self.norms.append(
                        nn.LayerNorm(
                            [in_channels, int(self.input_size[1] / scale), int(self.input_size[0] / scale)],
                            elementwise_affine=True,
                        )
                )

        self.nf_flows = nn.ModuleList()

        for in_channels, scale in zip(channels, scales):
            self.nf_flows.append(
                nf_fast_flow(
                    [in_channels, int(input_size[1] / scale), int(input_size[0] / scale)],
                    conv3x3_only=conv3x3_only,
                    hidden_ratio=hidden_ratio,
                    flow_steps=flow_steps,
                )
            )


    def forward(self, features):
        if self.backbone_name.find('resnet') != -1:
            features = [self.norms[i](feature) for i, feature in enumerate(features)]

        loss = 0
        outputs = []
        for i, feature in enumerate(features):
            output, log_jac_dets = self.nf_flows[i](feature)
            loss += torch.mean(
                0.5 * torch.sum(output**2, dim=(1, 2, 3)) - log_jac_dets
            )
            outputs.append(output)
        ret = {"loss": loss}

        if not self.training:
            anomaly_map_list = []
            for output in outputs:
                log_prob = -torch.mean(output**2, dim=1, keepdim=True) * 0.5

                prob = torch.exp(log_prob)
                a_map = F.interpolate(
                    -prob,
                    size=[self.input_size[1], self.input_size[0]],
                    mode="bilinear",
                    align_corners=False,
                )

                anomaly_map_list.append(a_map)
            anomaly_map_list = torch.stack(anomaly_map_list, dim=-1)
            anomaly_map = torch.mean(anomaly_map_list, dim=-1)
            ret["anomaly_map"] = anomaly_map
        return ret
